"""
Dataset Extraction API
"""

import os
import json
import analysis_engine.iex.extract_df_from_redis as iex_extract_utils
import analysis_engine.yahoo.extract_df_from_redis as yahoo_extract_utils
from analysis_engine.consts import SUCCESS
from analysis_engine.consts import FAILED
from analysis_engine.utils import get_last_close_str
from analysis_engine.api_requests import get_ds_dict
from spylunking.log.setup_logging import build_colorized_logger


log = build_colorized_logger(
    name=__name__)


def extract(
        ticker=None,
        tickers=None,
        use_key=None,
        extract_mode='all',
        iex_datasets=None,
        redis_enabled=True,
        redis_address=None,
        redis_db=None,
        redis_password=None,
        redis_expire=None,
        s3_enabled=True,
        s3_address=None,
        s3_bucket=None,
        s3_access_key=None,
        s3_secret_key=None,
        s3_region_name=None,
        s3_secure=False,
        celery_disabled=True,
        broker_url=None,
        result_backend=None,
        label=None,
        verbose=False):
    """extract

    Extract all cached datasets for a stock ``ticker`` or
    a list of ``tickers`` and returns a dictionary. Please
    make sure the datasets are already cached in Redis
    before running this method. If not please refer to
    the ``analysis_engine.fetch.fetch`` function
    to prepare the datasets on your environment.

    Python example:

    .. code-block:: python

        from analysis_engine.extract import extract
        d = extract(ticker='NFLX')
        print(d)
        for k in d['NFLX']:
            print('dataset key: {}'.format(k))


    This was created for reducing the amount of typying in
    Jupyter notebooks. It can be set up for use with a
    distributed engine as well with the optional arguments
    depending on your connectitivty requirements.

    .. note:: Please ensure Redis and Minio are running
              before trying to extract tickers

    **Stock tickers to extract**

    :param ticker: single stock ticker/symbol/ETF to extract
    :param tickers: optional - list of tickers to extract
    :param use_key: optional - extract historical key from Redis
        usually formatted ``<TICKER>_<date formatted YYYY-MM-DD>``

    **(Optional) Data sources, datafeeds and datasets to gather**

    :param iex_datasets: list of strings for gathering specific `IEX
        datasets <https://iextrading.com/developer/docs/#stocks>`__
        which are set as consts: ``analysis_engine.iex.consts.FETCH_*``.

    **(Optional) Redis connectivity arguments**

    :param redis_enabled: bool - toggle for auto-caching all
        datasets in Redis
        (default is ``True``)
    :param redis_address: Redis connection string format: ``host:port``
        (default is ``localhost:6379``)
    :param redis_db: Redis db to use
        (default is ``0``)
    :param redis_password: optional - Redis password
        (default is ``None``)
    :param redis_expire: optional - Redis expire value
        (default is ``None``)

    **(Optional) Minio (S3) connectivity arguments**

    :param s3_enabled: bool - toggle for auto-archiving on Minio (S3)
        (default is ``True``)
    :param s3_address: Minio S3 connection string format: ``host:port``
        (default is ``localhost:9000``)
    :param s3_bucket: S3 Bucket for storing the artifacts
        (default is ``dev``) which should be viewable on a browser:
        http://localhost:9000/minio/dev/
    :param s3_access_key: S3 Access key
        (default is ``trexaccesskey``)
    :param s3_secret_key: S3 Secret key
        (default is ``trex123321``)
    :param s3_region_name: S3 region name
        (default is ``us-east-1``)
    :param s3_secure: Transmit using tls encryption
        (default is ``False``)

    **(Optional) Celery worker broker connectivity arguments**

    :param celery_disabled: bool - toggle synchronous mode or publish
        to an engine connected to the `Celery broker and backend
        <https://github.com/celery/celery#transports-and-backends>`__
        (default is ``True`` - synchronous mode without an engine
        or need for a broker or backend for Celery)
    :param broker_url: Celery broker url
        (default is ``redis://0.0.0.0:6379/13``)
    :param result_backend: Celery backend url
        (default is ``redis://0.0.0.0:6379/14``)
    :param label: tracking log label

    **(Optional) Debugging**

    :param verbose: bool - show extract warnings
        and other debug logging (default is False)

    **Supported environment variables**

    ::

        export REDIS_ADDRESS="localhost:6379"
        export REDIS_DB="0"
        export S3_ADDRESS="localhost:9000"
        export S3_BUCKET="dev"
        export AWS_ACCESS_KEY_ID="trexaccesskey"
        export AWS_SECRET_ACCESS_KEY="trex123321"
        export AWS_DEFAULT_REGION="us-east-1"
        export S3_SECURE="0"
        export WORKER_BROKER_URL="redis://0.0.0.0:6379/13"
        export WORKER_BACKEND_URL="redis://0.0.0.0:6379/14"
    """

    rec = {}
    extract_requests = []

    use_tickers = tickers
    if ticker:
        use_tickers = [ticker]
    else:
        if not use_tickers:
            use_tickers = []

    default_iex_datasets = [
        'daily',
        'minute',
        'quote',
        'stats',
        'peers',
        'news',
        'financials',
        'earnings',
        'dividends',
        'company'
    ]

    if not iex_datasets:
        iex_datasets = default_iex_datasets

    if redis_enabled:
        if not redis_address:
            redis_address = os.getenv(
                'REDIS_ADDRESS',
                'localhost:6379')
        if not redis_password:
            redis_password = os.getenv(
                'REDIS_PASSWORD',
                None)
        if not redis_db:
            redis_db = int(os.getenv(
                'REDIS_DB',
                '0'))
        if not redis_expire:
            redis_expire = os.getenv(
                'REDIS_EXPIRE',
                None)
    if s3_enabled:
        if not s3_address:
            s3_address = os.getenv(
                'S3_ADDRESS',
                'localhost:9000')
        if not s3_access_key:
            s3_access_key = os.getenv(
                'AWS_ACCESS_KEY_ID',
                'trexaccesskey')
        if not s3_secret_key:
            s3_secret_key = os.getenv(
                'AWS_SECRET_ACCESS_KEY',
                'trex123321')
        if not s3_region_name:
            s3_region_name = os.getenv(
                'AWS_DEFAULT_REGION',
                'us-east-1')
        if not s3_secure:
            s3_secure = os.getenv(
                'S3_SECURE',
                '0') == '1'
        if not s3_bucket:
            s3_bucket = os.getenv(
                'S3_BUCKET',
                'dev')
    if not broker_url:
        broker_url = os.getenv(
            'WORKER_BROKER_URL',
            'redis://0.0.0.0:6379/13')
    if not result_backend:
        result_backend = os.getenv(
            'WORKER_BACKEND_URL',
            'redis://0.0.0.0:6379/14')

    if not label:
        label = 'get-latest'

    num_tickers = len(use_tickers)
    last_close_str = get_last_close_str()

    if iex_datasets:
        log.info(
            '{} - getting latest for tickers={} '
            'iex={}'.format(
                label,
                num_tickers,
                json.dumps(iex_datasets)))
    else:
        log.info(
            '{} - getting latest for tickers={}'.format(
                label,
                num_tickers))

    ticker_key = use_key
    if not ticker_key:
        ticker_key = '{}_{}'.format(
            ticker,
            last_close_str)

    common_vals = {}
    common_vals['base_key'] = ticker_key
    common_vals['celery_disabled'] = celery_disabled
    common_vals['ticker'] = ticker
    common_vals['label'] = label
    common_vals['iex_datasets'] = iex_datasets
    common_vals['s3_enabled'] = s3_enabled
    common_vals['s3_bucket'] = s3_bucket
    common_vals['s3_address'] = s3_address
    common_vals['s3_secure'] = s3_secure
    common_vals['s3_region_name'] = s3_region_name
    common_vals['s3_access_key'] = s3_access_key
    common_vals['s3_secret_key'] = s3_secret_key
    common_vals['s3_key'] = ticker_key
    common_vals['redis_enabled'] = redis_enabled
    common_vals['redis_address'] = redis_address
    common_vals['redis_password'] = redis_password
    common_vals['redis_db'] = redis_db
    common_vals['redis_key'] = ticker_key
    common_vals['redis_expire'] = redis_expire

    common_vals['redis_address'] = redis_address
    common_vals['s3_address'] = s3_address

    log.info(
        '{} - extract ticker={} last_close={} base_key={} '
        'redis_address={} s3_address={}'.format(
            label,
            ticker,
            last_close_str,
            common_vals['base_key'],
            common_vals['redis_address'],
            common_vals['s3_address']))

    """
    Extract Datasets
    """

    iex_daily_status = FAILED
    iex_minute_status = FAILED
    iex_quote_status = FAILED
    iex_stats_status = FAILED
    iex_peers_status = FAILED
    iex_news_status = FAILED
    iex_financials_status = FAILED
    iex_earnings_status = FAILED
    iex_dividends_status = FAILED
    iex_company_status = FAILED
    yahoo_news_status = FAILED
    yahoo_options_status = FAILED
    yahoo_pricing_status = FAILED

    iex_daily_df = None
    iex_minute_df = None
    iex_quote_df = None
    iex_stats_df = None
    iex_peers_df = None
    iex_news_df = None
    iex_financials_df = None
    iex_earnings_df = None
    iex_dividends_df = None
    iex_company_df = None
    yahoo_option_calls_df = None
    yahoo_option_puts_df = None
    yahoo_pricing_df = None
    yahoo_news_df = None

    for ticker in use_tickers:
        req = get_ds_dict(
            ticker=ticker,
            base_key=common_vals['base_key'],
            ds_id=label,
            service_dict=common_vals)
        extract_requests.append(req)
    # end of for all ticker in use_tickers

    extract_iex = True
    if extract_mode not in ['all', 'iex']:
        extract_iex = False

    extract_yahoo = True
    if extract_mode not in ['all', 'yahoo']:
        extract_yahoo = False

    for extract_req in extract_requests:
        if 'daily' in iex_datasets or extract_iex:
            iex_daily_status, iex_daily_df = \
                iex_extract_utils.extract_daily_dataset(
                    extract_req)
            if iex_daily_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract iex_daily={}'.format(ticker))
        if 'minute' in iex_datasets or extract_iex:
            iex_minute_status, iex_minute_df = \
                iex_extract_utils.extract_minute_dataset(
                    extract_req)
            if iex_minute_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract iex_minute={}'.format(ticker))
        if 'quote' in iex_datasets or extract_iex:
            iex_quote_status, iex_quote_df = \
                iex_extract_utils.extract_quote_dataset(
                    extract_req)
            if iex_quote_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract iex_quote={}'.format(ticker))
        if 'stats' in iex_datasets or extract_iex:
            iex_stats_df, iex_stats_df = \
                iex_extract_utils.extract_stats_dataset(
                    extract_req)
            if iex_stats_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract iex_stats={}'.format(ticker))
        if 'peers' in iex_datasets or extract_iex:
            iex_peers_df, iex_peers_df = \
                iex_extract_utils.extract_peers_dataset(
                    extract_req)
            if iex_peers_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract iex_peers={}'.format(ticker))
        if 'news' in iex_datasets or extract_iex:
            iex_news_status, iex_news_df = \
                iex_extract_utils.extract_news_dataset(
                    extract_req)
            if iex_news_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract iex_news={}'.format(ticker))
        if 'financials' in iex_datasets or extract_iex:
            iex_financials_status, iex_financials_df = \
                iex_extract_utils.extract_financials_dataset(
                    extract_req)
            if iex_financials_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract iex_financials={}'.format(ticker))
        if 'earnings' in iex_datasets or extract_iex:
            iex_earnings_status, iex_earnings_df = \
                iex_extract_utils.extract_earnings_dataset(
                    extract_req)
            if iex_earnings_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract iex_earnings={}'.format(ticker))
        if 'dividends' in iex_datasets or extract_iex:
            iex_dividends_status, iex_dividends_df = \
                iex_extract_utils.extract_dividends_dataset(
                    extract_req)
            if iex_dividends_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract iex_dividends={}'.format(ticker))
        if 'company' in iex_datasets or extract_iex:
            iex_company_status, iex_company_df = \
                iex_extract_utils.extract_company_dataset(
                    extract_req)
            if iex_company_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract iex_company={}'.format(ticker))
        # end of iex extracts

        if extract_yahoo:
            yahoo_options_status, yahoo_option_calls_df = \
                yahoo_extract_utils.extract_option_calls_dataset(
                    extract_req)
            yahoo_options_status, yahoo_option_puts_df = \
                yahoo_extract_utils.extract_option_puts_dataset(
                    extract_req)
            if yahoo_options_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract yahoo_options={}'.format(ticker))
            yahoo_pricing_status, yahoo_pricing_df = \
                yahoo_extract_utils.extract_pricing_dataset(
                    extract_req)
            if yahoo_pricing_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract yahoo_pricing={}'.format(ticker))
            yahoo_news_status, yahoo_news_df = \
                yahoo_extract_utils.extract_yahoo_news_dataset(
                    extract_req)
            if yahoo_news_status != SUCCESS:
                if verbose:
                    log.warning(
                        'unable to extract yahoo_news={}'.format(ticker))
        # end of yahoo extracts

        ticker_data = {}
        ticker_data['daily'] = iex_daily_df
        ticker_data['minute'] = iex_minute_df
        ticker_data['quote'] = iex_quote_df
        ticker_data['stats'] = iex_stats_df
        ticker_data['peers'] = iex_peers_df
        ticker_data['news1'] = iex_news_df
        ticker_data['financials'] = iex_financials_df
        ticker_data['earnings'] = iex_earnings_df
        ticker_data['dividends'] = iex_dividends_df
        ticker_data['company'] = iex_company_df
        ticker_data['calls'] = yahoo_option_calls_df
        ticker_data['puts'] = yahoo_option_puts_df
        ticker_data['pricing'] = yahoo_pricing_df
        ticker_data['news'] = yahoo_news_df

        rec[ticker] = ticker_data
    # end of for service_dict in extract_requests

    return rec
# end of extract



"""
Mock Algorithm Methods for unittesting things
like previously-owned shares or sell-side indicators
without owning shares
"""

import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def mock_algo_owns_shares_in_ticker_before_starting(
        obj,
        ticker):
    """mock_algo_owns_shares_in_ticker_before_starting

    Support mocking owned shares to test indicator selling

    If you can modify your algorithm ``config_dict`` you can
    also set a ``positions`` dictionary like:

    .. code-block:: python

        algo_config_dict = {
            # other values omitted for docs
            'positions': {
                'SPY': {
                    'shares': 10000,
                    'buys': [],
                    'sells': []
                }
            }
        }


    Use with your custom algorithm unittests:

    .. code-block:: python

        import mock
        import analysis_engine.mocks.mock_algo_trading as mock_trading

        @mock.patch(
            ('analysis_engine.algo.BaseAlgo.get_ticker_positions'),
            new=mock_trading.mock_algo_owns_shares_in_ticker_before_starting)


    :param obj: algorithm object
    :param ticker: ticker symbol
    """
    num_owned = 10000
    buys = []
    sells = []
    return num_owned, buys, sells
# end of mock_algo_owns_shares_in_ticker_before_starting



"""
Convert json to a pandas dataframe
"""

import pandas as pd
from spylunking.log.setup_logging import build_colorized_logger

log = build_colorized_logger(
    name=__name__)


def convert_json_to_df(
        data_json,
        sorted_by_key="date",
        in_ascending=True):
    """convert_json_to_df

    Convert json to a pandas dataframe

    :param data_json: json to convert
    :param sorted_by_key: key to sort on
    :param in_ascending: ascending order (True by default)
    """
    log.info(
        'start sort={} asc={}'.format(
            sorted_by_key,
            in_ascending))
    new_df = pd.read_json(
        data_json).sort_values(
            by=sorted_by_key,
            ascending=in_ascending)
    log.info(
        'done')
    return new_df
# end of convert_json_to_df



"""
This is a wrapper for running your own custom algorithms

.. note:: Please refer to the `sa.py <https://github.com/Algo
    Traders/stock-analysis-engine/blob/master/analysi
    s_engine/scripts/sa.py>`__ for the lastest usage examples.

Example with the command line tool:

::

    bt -t SPY -g /opt/sa/analysis_engine/mocks/example_algo_minute.py

"""

import os
import inspect
import types
import importlib.machinery
import datetime
import json
import analysis_engine.consts as ae_consts
import analysis_engine.build_algo_request as build_algo_request
import analysis_engine.build_publish_request as build_publish_request
import analysis_engine.build_result as build_result
import analysis_engine.run_algo as run_algo
import analysis_engine.work_tasks.get_celery_app as get_celery_app
import analysis_engine.algo as ae_algo
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def run_custom_algo(
        mod_path,
        ticker='SPY',
        balance=50000,
        commission=6.0,
        start_date=None,
        end_date=None,
        name='myalgo',
        auto_fill=True,
        config_file=None,
        config_dict=None,
        load_from_s3_bucket=None,
        load_from_s3_key=None,
        load_from_redis_key=None,
        load_from_file=None,
        load_compress=False,
        load_publish=True,
        load_config=None,
        report_redis_key=None,
        report_s3_bucket=None,
        report_s3_key=None,
        report_file=None,
        report_compress=False,
        report_publish=True,
        report_config=None,
        history_redis_key=None,
        history_s3_bucket=None,
        history_s3_key=None,
        history_file=None,
        history_compress=False,
        history_publish=True,
        history_config=None,
        extract_redis_key=None,
        extract_s3_bucket=None,
        extract_s3_key=None,
        extract_file=None,
        extract_save_dir=None,
        extract_compress=False,
        extract_publish=True,
        extract_config=None,
        publish_to_s3=True,
        publish_to_redis=True,
        publish_to_slack=True,
        dataset_type=ae_consts.SA_DATASET_TYPE_ALGO_READY,
        serialize_datasets=ae_consts.DEFAULT_SERIALIZED_DATASETS,
        compress=False,
        encoding='utf-8',
        redis_enabled=True,
        redis_key=None,
        redis_address=None,
        redis_db=None,
        redis_password=None,
        redis_expire=None,
        redis_serializer='json',
        redis_encoding='utf-8',
        s3_enabled=True,
        s3_key=None,
        s3_address=None,
        s3_bucket=None,
        s3_access_key=None,
        s3_secret_key=None,
        s3_region_name=None,
        s3_secure=False,
        slack_enabled=False,
        slack_code_block=False,
        slack_full_width=False,
        timeseries=None,
        trade_strategy=None,
        verbose=False,
        debug=False,
        dataset_publish_extract=False,
        dataset_publish_history=False,
        dataset_publish_report=False,
        run_on_engine=False,
        auth_url=ae_consts.WORKER_BROKER_URL,
        backend_url=ae_consts.WORKER_BACKEND_URL,
        include_tasks=ae_consts.INCLUDE_TASKS,
        ssl_options=ae_consts.SSL_OPTIONS,
        transport_options=ae_consts.TRANSPORT_OPTIONS,
        path_to_config_module=ae_consts.WORKER_CELERY_CONFIG_MODULE,
        raise_on_err=True):
    """run_custom_algo

    Run a custom algorithm that derives the
    ``analysis_engine.algo.BaseAlgo`` class

    .. note:: Make sure to only have **1**
        class defined in an algo module. Imports from
        other modules should work just fine.

    **Algorithm arguments**

    :param mod_path: file path to custom
        algorithm class module
    :param ticker: ticker symbol
    :param balance: float - starting balance capital
        for creating buys and sells
    :param commission: float - cost pet buy or sell
    :param name: string - name for tracking algorithm
        in the logs
    :param start_date: string - start date for backtest with
        format ``YYYY-MM-DD HH:MM:SS``
    :param end_date: end date for backtest with
        format ``YYYY-MM-DD HH:MM:SS``
    :param auto_fill: optional - boolean for auto filling
        buy and sell orders for backtesting
        (default is ``True``)
    :param config_file: path to a json file
        containing custom algorithm object
        member values (like indicator configuration and
        predict future date units ahead for a backtest)
    :param config_dict: optional - dictionary that
        can be passed to derived class implementations
        of: ``def load_from_config(config_dict=config_dict)``

    **Timeseries**

    :param timeseries: optional - string to
        set ``day`` or ``minute`` backtesting
        or live trading
        (default is ``minute``)

    **Trading Strategy**

    :param trade_strategy: optional - string to
        set the type of ``Trading Strategy``
        for backtesting or live trading
        (default is ``count``)

    **Running Distributed Algorithms on the Engine Workers**

    :param run_on_engine: optional - boolean
        flag for publishing custom algorithms
        to Celery analysis_engine workers for distributing
        algorithm workloads
        (default is ``False`` which will run algos locally)
        this is required for distributing algorithms
    :param auth_url: Celery broker address
        (default is ``redis://localhost:6379/13``
        or ``analysis_engine.consts.WORKER_BROKER_URL``
        environment variable)
        this is required for distributing algorithms
    :param backend_url: Celery backend address
        (default is ``redis://localhost:6379/14``
        or ``analysis_engine.consts.WORKER_BACKEND_URL``
        environment variable)
        this is required for distributing algorithms
    :param include_tasks: list of modules containing tasks to add
        (default is ``analysis_engine.consts.INCLUDE_TASKS``)
    :param ssl_options: security options dictionary
        (default is ``analysis_engine.consts.SSL_OPTIONS``)
    :param trasport_options: transport options dictionary
        (default is ``analysis_engine.consts.TRANSPORT_OPTIONS``)
    :param path_to_config_module: config module for advanced
        Celery worker connectivity requirements
        (default is ``analysis_engine.work_tasks.celery_config``
        or ``analysis_engine.consts.WORKER_CELERY_CONFIG_MODULE``)

    **Load Algorithm-Ready Dataset From Source**

    Use these arguments to load algorithm-ready datasets
    from supported sources (file, s3 or redis)

    :param load_from_s3_bucket: optional - string load the algo from an
        a previously-created s3 bucket holding an s3 key with an
        algorithm-ready dataset for use with:
        ``handle_data``
    :param load_from_s3_key: optional - string load the algo from an
        a previously-created s3 key holding an
        algorithm-ready dataset for use with:
        ``handle_data``
    :param load_from_redis_key: optional - string load the algo from a
        a previously-created redis key holding an
        algorithm-ready dataset for use with:
        ``handle_data``
    :param load_from_file: optional - string path to
        a previously-created local file holding an
        algorithm-ready dataset for use with:
        ``handle_data``
    :param load_compress: optional - boolean
        flag for toggling to decompress
        or not when loading an algorithm-ready
        dataset (``True`` means the dataset
        must be decompressed to load correctly inside
        an algorithm to run a backtest)
    :param load_publish: boolean - toggle publishing
        the load progress to slack, s3, redis or a file
        (default is ``True``)
    :param load_config: optional - dictionary
        for setting member variables to load an
        agorithm-ready dataset from
        a file, s3 or redis

    **Publishing Control Bool Flags**

    :param publish_to_s3: optional - boolean for
        toggling publishing to s3 on/off
        (default is ``True``)
    :param publish_to_redis: optional - boolean for
        publishing to redis on/off
        (default is ``True``)
    :param publish_to_slack: optional - boolean for
        publishing to slack
        (default is ``True``)

    **Algorithm Trade History Arguments**

    :param history_redis_key: optional - string
        where the algorithm trading history will be stored in
        an redis key
    :param history_s3_bucket: optional - string
        where the algorithm trading history will be stored in
        an s3 bucket
    :param history_s3_key: optional - string
        where the algorithm trading history will be stored in
        an s3 key
    :param history_file: optional - string key
        where the algorithm trading history will be stored in
        a file serialized as a json-string
    :param history_compress: optional - boolean
        flag for toggling to decompress
        or not when loading an algorithm-ready
        dataset (``True`` means the dataset
        will be compressed on publish)
    :param history_publish: boolean - toggle publishing
        the history to s3, redis or a file
        (default is ``True``)
    :param history_config: optional - dictionary
        for setting member variables to publish
        an algo ``trade history`` to s3, redis, a file
        or slack

    **Algorithm Trade Performance Report Arguments (Output Dataset)**

    :param report_redis_key: optional - string
        where the algorithm ``trading performance report`` (report)
        will be stored in an redis key
    :param report_s3_bucket: optional - string
        where the algorithm report will be stored in
        an s3 bucket
    :param report_s3_key: optional - string
        where the algorithm report will be stored in
        an s3 key
    :param report_file: optional - string key
        where the algorithm report will be stored in
        a file serialized as a json-string
    :param report_compress: optional - boolean
        flag for toggling to decompress
        or not when loading an algorithm-ready
        dataset (``True`` means the dataset
        will be compressed on publish)
    :param report_publish: boolean - toggle publishing
        the ``trading performance report`` s3, redis or a file
        (default is ``True``)
    :param report_config: optional - dictionary
        for setting member variables to publish
        an algo ``trading performance report`` to s3,
        redis, a file or slack

    **Extract an Algorithm-Ready Dataset Arguments**

    :param extract_redis_key: optional - string
        where the algorithm report will be stored in
        an redis key
    :param extract_s3_bucket: optional - string
        where the algorithm report will be stored in
        an s3 bucket
    :param extract_s3_key: optional - string
        where the algorithm report will be stored in
        an s3 key
    :param extract_file: optional - string key
        where the algorithm report will be stored in
        a file serialized as a json-string
    :param extract_save_dir: optional - string path to
        auto-generated files from the algo
    :param extract_compress: optional - boolean
        flag for toggling to decompress
        or not when loading an algorithm-ready
        dataset (``True`` means the dataset
        will be compressed on publish)
    :param extract_publish: boolean - toggle publishing
        the used ``algorithm-ready dataset`` to s3, redis or a file
        (default is ``True``)
    :param extract_config: optional - dictionary
        for setting member variables to publish
        an algo ``trading performance report`` to s3,
        redis, a file or slack

    **Dataset Arguments**

    :param dataset_type: optional - dataset type
        (default is ``SA_DATASET_TYPE_ALGO_READY``)
    :param serialize_datasets: optional - list of dataset names to
        deserialize in the dataset
        (default is ``DEFAULT_SERIALIZED_DATASETS``)
    :param encoding: optional - string for data encoding

    **Publish Algorithm Datasets to S3, Redis or a File**

    :param dataset_publish_extract: optional - bool
        for publishing the algorithm's
        ``algorithm-ready``
        dataset to: s3, redis or file
    :param dataset_publish_history: optional - bool
        for publishing the algorithm's
        ``trading history``
        dataset to: s3, redis or file
    :param dataset_publish_report: optional - bool
        for publishing the algorithm's
        ``trading performance report``
        dataset to: s3, redis or file

    **Redis connectivity arguments**

    :param redis_enabled: bool - toggle for auto-caching all
        datasets in Redis
        (default is ``True``)
    :param redis_key: string - key to save the data in redis
        (default is ``None``)
    :param redis_address: Redis connection string format: ``host:port``
        (default is ``localhost:6379``)
    :param redis_db: Redis db to use
        (default is ``0``)
    :param redis_password: optional - Redis password
        (default is ``None``)
    :param redis_expire: optional - Redis expire value
        (default is ``None``)
    :param redis_serializer: not used yet - support for future
        pickle objects in redis
    :param redis_encoding: format of the encoded key in redis

    **Minio (S3) connectivity arguments**

    :param s3_enabled: bool - toggle for auto-archiving on Minio (S3)
        (default is ``True``)
    :param s3_key: string - key to save the data in redis
        (default is ``None``)
    :param s3_address: Minio S3 connection string format: ``host:port``
        (default is ``localhost:9000``)
    :param s3_bucket: S3 Bucket for storing the artifacts
        (default is ``dev``) which should be viewable on a browser:
        http://localhost:9000/minio/dev/
    :param s3_access_key: S3 Access key
        (default is ``trexaccesskey``)
    :param s3_secret_key: S3 Secret key
        (default is ``trex123321``)
    :param s3_region_name: S3 region name
        (default is ``us-east-1``)
    :param s3_secure: Transmit using tls encryption
        (default is ``False``)

    **Slack arguments**

    :param slack_enabled: optional - boolean for
        publishing to slack
    :param slack_code_block: optional - boolean for
        publishing as a code black in slack
    :param slack_full_width: optional - boolean for
        publishing as a to slack using the full
        width allowed

    **Debugging arguments**

    :param debug: optional - bool for debug tracking
    :param verbose: optional - bool for increasing
        logging
    :param raise_on_err: boolean - set this to ``False`` on prod
        to ensure exceptions do not interrupt services.
        With the default (``True``) any exceptions from the library
        and your own algorithm are sent back out immediately exiting
        the backtest.
    """

    module_name = 'BaseAlgo'
    custom_algo_module = None
    new_algo_object = None
    use_custom_algo = False
    found_algo_module = True
    should_publish_extract_dataset = False
    should_publish_history_dataset = False
    should_publish_report_dataset = False
    use_config_file = None
    use_config_dict = config_dict
    if config_file:
        if os.path.exists(config_file):
            use_config_file = config_file
            if not config_dict:
                try:
                    use_config_dict = json.loads(open(
                        config_file, 'r').read())
                except Exception as e:
                    msg = (
                        'failed parsing json config_file={} '
                        'with ex={}'.format(
                            config_file,
                            e))
                    log.error(msg)
                    raise Exception(msg)
    # end of loading the config_file

    err = None
    if mod_path:
        module_name = mod_path.split('/')[-1]
        loader = importlib.machinery.SourceFileLoader(
            module_name,
            mod_path)
        custom_algo_module = types.ModuleType(
            loader.name)
        loader.exec_module(
            custom_algo_module)
        use_custom_algo = True

        for member in inspect.getmembers(custom_algo_module):
            if module_name in str(member):
                found_algo_module = True
                break
        # for all members in this custom module file
    # if loading a custom algorithm module from a file on disk

    if not found_algo_module:
        err = (
            'unable to find custom algorithm module={}'.format(
                custom_algo_module))
        if mod_path:
            err = (
                'analysis_engine.run_custom_algo.run_custom_algo was unable '
                'to find custom algorithm module={} with provided path to \n '
                'file: {} \n'
                '\n'
                'Please confirm '
                'that the class inherits from the BaseAlgo class like:\n'
                '\n'
                'import analysis_engine.algo\n'
                'class MyAlgo(analysis_engine.algo.BaseAlgo):\n '
                '\n'
                'If it is then please file an issue on github:\n '
                'https://github.com/AlgoTraders/stock-analysis-engine/'
                'issues/new \n\nFor now this error results in a shutdown'
                '\n'.format(
                    custom_algo_module,
                    mod_path))
        # if mod_path set

        if verbose or debug:
            log.error(err)
        return build_result.build_result(
            status=ae_consts.ERR,
            err=err,
            rec=None)
    # if not found_algo_module

    use_start_date = start_date
    use_end_date = end_date
    if not use_end_date:
        end_date = datetime.datetime.utcnow()
        use_end_date = end_date.strftime(
            ae_consts.COMMON_TICK_DATE_FORMAT)
    if not use_start_date:
        start_date = end_date - datetime.timedelta(days=75)
        use_start_date = start_date.strftime(
            ae_consts.COMMON_TICK_DATE_FORMAT)
        if verbose:
            log.info(
                '{} {} setting default start_date={}'.format(
                    name,
                    ticker,
                    use_start_date))

    # Load an algorithm-ready dataset from:
    # file, s3, or redis
    if not load_config:
        load_config = build_publish_request.build_publish_request(
            ticker=ticker,
            output_file=None,
            s3_bucket=None,
            s3_key=None,
            redis_key=None,
            compress=load_compress,
            redis_enabled=publish_to_redis,
            redis_address=redis_address,
            redis_db=redis_db,
            redis_password=redis_password,
            redis_expire=redis_expire,
            redis_serializer=redis_serializer,
            redis_encoding=redis_encoding,
            s3_enabled=publish_to_s3,
            s3_address=s3_address,
            s3_access_key=s3_access_key,
            s3_secret_key=s3_secret_key,
            s3_region_name=s3_region_name,
            s3_secure=s3_secure,
            slack_enabled=publish_to_slack,
            slack_code_block=slack_code_block,
            slack_full_width=slack_full_width,
            verbose=verbose,
            label='load-{}'.format(name))
        if load_from_file:
            load_config['output_file'] = load_from_file
        if load_from_redis_key:
            load_config['redis_key'] = load_from_redis_key
            load_config['redis_enabled'] = True
        if load_from_s3_bucket and load_from_s3_key:
            load_config['s3_bucket'] = load_from_s3_bucket
            load_config['s3_key'] = load_from_s3_key
            load_config['s3_enabled'] = True
    # end of building load_config dictionary if not already set

    # Automatically save all datasets to an algorithm-ready:
    # file, s3, or redis
    if not extract_config:
        extract_config = build_publish_request.build_publish_request(
            ticker=ticker,
            output_file=None,
            s3_bucket=None,
            s3_key=None,
            redis_key=None,
            compress=extract_compress,
            redis_enabled=publish_to_redis,
            redis_address=redis_address,
            redis_db=redis_db,
            redis_password=redis_password,
            redis_expire=redis_expire,
            redis_serializer=redis_serializer,
            redis_encoding=redis_encoding,
            s3_enabled=publish_to_s3,
            s3_address=s3_address,
            s3_access_key=s3_access_key,
            s3_secret_key=s3_secret_key,
            s3_region_name=s3_region_name,
            s3_secure=s3_secure,
            slack_enabled=publish_to_slack,
            slack_code_block=slack_code_block,
            slack_full_width=slack_full_width,
            verbose=verbose,
            label='extract-{}'.format(name))
        should_publish_extract_dataset = False
        if extract_file:
            extract_config['output_file'] = extract_file
            should_publish_extract_dataset = True
        if extract_redis_key and publish_to_redis:
            extract_config['redis_key'] = extract_redis_key
            extract_config['redis_enabled'] = True
            should_publish_extract_dataset = True
        if extract_s3_bucket and extract_s3_key and publish_to_s3:
            extract_config['s3_bucket'] = extract_s3_bucket
            extract_config['s3_key'] = extract_s3_key
            extract_config['s3_enabled'] = True
            should_publish_extract_dataset = True
    # end of building extract_config dictionary if not already set

    # Automatically save the trading performance report:
    # file, s3, or redis
    if not report_config:
        report_config = build_publish_request.build_publish_request(
            ticker=ticker,
            output_file=None,
            s3_bucket=None,
            s3_key=None,
            redis_key=None,
            compress=report_compress,
            redis_enabled=publish_to_redis,
            redis_address=redis_address,
            redis_db=redis_db,
            redis_password=redis_password,
            redis_expire=redis_expire,
            redis_serializer=redis_serializer,
            redis_encoding=redis_encoding,
            s3_enabled=publish_to_s3,
            s3_address=s3_address,
            s3_access_key=s3_access_key,
            s3_secret_key=s3_secret_key,
            s3_region_name=s3_region_name,
            s3_secure=s3_secure,
            slack_enabled=publish_to_slack,
            slack_code_block=slack_code_block,
            slack_full_width=slack_full_width,
            verbose=verbose,
            label='report-{}'.format(name))
        should_publish_report_dataset = False
        if report_file:
            report_config['output_file'] = report_file
            should_publish_report_dataset = True
        if report_redis_key and publish_to_redis:
            report_config['redis_key'] = report_redis_key
            report_config['redis_enabled'] = True
            should_publish_report_dataset = True
        if report_s3_bucket and report_s3_key and publish_to_s3:
            report_config['s3_bucket'] = report_s3_bucket
            report_config['s3_key'] = report_s3_key
            report_config['s3_enabled'] = True
            should_publish_report_dataset = True
    # end of building report_config dictionary if not already set

    # Automatically save the trade history:
    # file, s3, or redis
    if not history_config:
        history_config = build_publish_request.build_publish_request(
            ticker=ticker,
            output_file=None,
            s3_bucket=None,
            s3_key=None,
            redis_key=None,
            compress=report_compress,
            redis_enabled=publish_to_redis,
            redis_address=redis_address,
            redis_db=redis_db,
            redis_password=redis_password,
            redis_expire=redis_expire,
            redis_serializer=redis_serializer,
            redis_encoding=redis_encoding,
            s3_enabled=publish_to_s3,
            s3_address=s3_address,
            s3_access_key=s3_access_key,
            s3_secret_key=s3_secret_key,
            s3_region_name=s3_region_name,
            s3_secure=s3_secure,
            slack_enabled=publish_to_slack,
            slack_code_block=slack_code_block,
            slack_full_width=slack_full_width,
            verbose=verbose,
            label='history-{}'.format(name))
        should_publish_history_dataset = False
        if history_file:
            history_config['output_file'] = history_file
            should_publish_history_dataset = True
        if history_redis_key and publish_to_redis:
            history_config['redis_key'] = history_redis_key
            history_config['redis_enabled'] = True
            should_publish_history_dataset = True
        if history_s3_bucket and history_s3_key and publish_to_s3:
            history_config['s3_bucket'] = history_s3_bucket
            history_config['s3_key'] = history_s3_key
            history_config['s3_enabled'] = True
            should_publish_history_dataset = True
    # end of building history_config dictionary if not already set

    if verbose:
        remove_vals = [
            's3_access_key',
            's3_secret_key',
            'redis_password'
        ]
        debug_extract_config = {}
        for k in extract_config:
            if k not in remove_vals:
                debug_extract_config[k] = extract_config[k]
        debug_report_config = {}
        for k in report_config:
            if k not in remove_vals:
                debug_report_config[k] = report_config[k]
        debug_history_config = {}
        for k in history_config:
            if k not in remove_vals:
                debug_history_config[k] = history_config[k]
        debug_load_config = {}
        for k in load_config:
            if k not in remove_vals:
                debug_load_config[k] = load_config[k]
        log.info(
            '{} {} using extract config {}'.format(
                name,
                ticker,
                ae_consts.ppj(debug_extract_config)))
        log.info(
            '{} {} using report config {}'.format(
                name,
                ticker,
                ae_consts.ppj(debug_report_config)))
        log.info(
            '{} {} using trade history config {}'.format(
                name,
                ticker,
                ae_consts.ppj(debug_history_config)))
        log.info(
            '{} {} using load config {}'.format(
                name,
                ticker,
                ae_consts.ppj(debug_load_config)))
        log.info(
            '{} {} - building algo request'.format(
                name,
                ticker))
    # end of verbose

    algo_req = build_algo_request.build_algo_request(
        ticker=ticker,
        balance=balance,
        commission=commission,
        start_date=use_start_date,
        end_date=use_end_date,
        timeseries=timeseries,
        trade_strategy=trade_strategy,
        config_file=use_config_file,
        config_dict=use_config_dict,
        load_config=load_config,
        history_config=history_config,
        report_config=report_config,
        extract_config=extract_config,
        label=name)

    algo_req['name'] = name

    algo_res = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=None)

    if run_on_engine:
        rec = {
            'algo_req': algo_req,
            'task_id': None
        }
        task_name = (
            'analysis_engine.work_tasks.'
            'run_distributed_algorithm.run_distributed_algorithm')
        if verbose:
            log.info(
                'starting distributed algo task={}'.format(
                    task_name))
        elif debug:
            log.info(
                'starting distributed algo by publishing to '
                'task={} broker={} backend={}'.format(
                    task_name,
                    auth_url,
                    backend_url))

        # Get the Celery app
        app = get_celery_app.get_celery_app(
            name=__name__,
            auth_url=auth_url,
            backend_url=backend_url,
            path_to_config_module=path_to_config_module,
            ssl_options=ssl_options,
            transport_options=transport_options,
            include_tasks=include_tasks)

        if debug:
            log.info(
                'calling distributed algo task={} request={}'.format(
                    task_name,
                    ae_consts.ppj(algo_req)))
        elif verbose:
            log.info(
                'calling distributed algo task={}'.format(
                    task_name))

        job_id = app.send_task(
            task_name,
            (algo_req,))
        if verbose:
            log.info(
                'calling task={} - success job_id={}'.format(
                    task_name,
                    job_id))
        rec['task_id'] = job_id
        algo_res = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)
        return algo_res
    # end of run_on_engine

    if use_custom_algo:
        if verbose:
            log.info(
                'inspecting {} for class {}'.format(
                    custom_algo_module,
                    module_name))
        use_class_member_object = None
        for member in inspect.getmembers(custom_algo_module):
            if module_name in str(member):
                if verbose:
                    log.info(
                        'start {} with {}'.format(
                            name,
                            member[1]))
                use_class_member_object = member
                break
        # end of looking over the class definition but did not find it

        if use_class_member_object:
            new_algo_object = member[1](
                **algo_req)
        else:
            err = (
                'did not find a derived analysis_engine.algo.BaseAlgo '
                'class in the module file={} '
                'for ticker={} algo_name={}'.format(
                    mod_path,
                    ticker,
                    name))

            if verbose or debug:
                log.error(err)

            return build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=None)
        # end of finding a valid algorithm object
    else:
        new_algo_object = ae_algo.BaseAlgo(
            **algo_req)
    # if using a custom module path or the BaseAlgo

    if new_algo_object:
        # heads up - logging this might have passwords in the algo_req
        # log.debug(
        #     '{} algorithm request: {}'.format(
        #         name,
        #         algo_req))
        if verbose:
            log.info(
                '{} - run ticker={} from {} to {}'.format(
                    name,
                    ticker,
                    use_start_date,
                    use_end_date))
        algo_res = run_algo.run_algo(
            algo=new_algo_object,
            raise_on_err=raise_on_err,
            **algo_req)
        algo_res['algo'] = new_algo_object
        if verbose:
            log.info(
                '{} - run ticker={} from {} to {}'.format(
                    name,
                    ticker,
                    use_start_date,
                    use_end_date))
        if verbose or debug:
            if custom_algo_module:
                log.info(
                    '{} - done run_algo custom_algo_module={} '
                    'module_name={} '
                    'ticker={} from {} to {}'.format(
                        name,
                        custom_algo_module,
                        module_name,
                        ticker,
                        use_start_date,
                        use_end_date))
            else:
                log.info(
                    '{} - done run_algo BaseAlgo ticker={} '
                    'from {} to {}'.format(
                        name,
                        ticker,
                        use_start_date,
                        use_end_date))
    else:
        err = (
            'missing a derived analysis_engine.algo.BaseAlgo '
            'class in the module file={} for ticker={} algo_name={}'.format(
                mod_path,
                ticker,
                name))
        return build_result.build_result(
            status=ae_consts.ERR,
            err=err,
            rec=None)
    # end of finding a valid algorithm object

    algo = algo_res.get(
        'algo',
        None)

    if not algo:
        err = (
            'failed creating algorithm object - '
            'ticker={} status={} error={}'
            'algo name={} custom_algo_module={} module_name={} '
            'from {} to {}'.format(
                ticker,
                ae_consts.get_status(status=algo_res['status']),
                algo_res['err'],
                name,
                custom_algo_module,
                module_name,
                use_start_date,
                use_end_date))
        return build_result.build_result(
            status=ae_consts.ERR,
            err=err,
            rec=None)

    if should_publish_extract_dataset and dataset_publish_extract:
        s3_log = ''
        redis_log = ''
        file_log = ''
        use_log = 'publish'

        if (extract_config['redis_address'] and
                extract_config['redis_db'] and
                extract_config['redis_key']):
            redis_log = 'redis://{}@{}/{}'.format(
                extract_config['redis_address'],
                extract_config['redis_db'],
                extract_config['redis_key'])
            use_log += ' {}'.format(
                redis_log)
        else:
            extract_config['redis_enabled'] = False
        if (extract_config['s3_address'] and
                extract_config['s3_bucket'] and
                extract_config['s3_key']):
            s3_log = 's3://{}/{}/{}'.format(
                extract_config['s3_address'],
                extract_config['s3_bucket'],
                extract_config['s3_key'])
            use_log += ' {}'.format(
                s3_log)
        else:
            extract_config['s3_enabled'] = False
        if extract_config['output_file']:
            file_log = 'file:{}'.format(
                extract_config['output_file'])
            use_log += ' {}'.format(
                file_log)

        if verbose:
            log.info(
                '{} - publish - start ticker={} algorithm-ready {}'
                ''.format(
                    name,
                    ticker,
                    use_log))

        publish_status = algo.publish_input_dataset(
            **extract_config)
        if publish_status != ae_consts.SUCCESS:
            msg = (
                'failed to publish algorithm-ready datasets '
                'with status {} attempted to {}'.format(
                    ae_consts.get_status(status=publish_status),
                    use_log))
            log.error(msg)
            return build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=None)

        if verbose:
            log.info(
                '{} - publish - done ticker={} algorithm-ready {}'
                ''.format(
                    name,
                    ticker,
                    use_log))
    # if publish the algorithm-ready dataset

    if should_publish_history_dataset and dataset_publish_history:
        s3_log = ''
        redis_log = ''
        file_log = ''
        use_log = 'publish'

        if (history_config['redis_address'] and
                history_config['redis_db'] and
                history_config['redis_key']):
            redis_log = 'redis://{}@{}/{}'.format(
                history_config['redis_address'],
                history_config['redis_db'],
                history_config['redis_key'])
            use_log += ' {}'.format(
                redis_log)
        else:
            history_config['redis_enabled'] = False
        if (history_config['s3_address'] and
                history_config['s3_bucket'] and
                history_config['s3_key']):
            s3_log = 's3://{}/{}/{}'.format(
                history_config['s3_address'],
                history_config['s3_bucket'],
                history_config['s3_key'])
            use_log += ' {}'.format(
                s3_log)
        else:
            history_config['s3_enabled'] = False
        if history_config['output_file']:
            file_log = 'file:{}'.format(
                history_config['output_file'])
            use_log += ' {}'.format(
                file_log)

        if verbose:
            log.info(
                '{} - publish - start ticker={} trading history {}'
                ''.format(
                    name,
                    ticker,
                    use_log))

        publish_status = algo.publish_trade_history_dataset(
            **history_config)
        if publish_status != ae_consts.SUCCESS:
            msg = (
                'failed to publish trading history datasets '
                'with status {} attempted to {}'.format(
                    ae_consts.get_status(status=publish_status),
                    use_log))
            log.error(msg)
            return build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=None)

        if verbose:
            log.info(
                '{} - publish - done ticker={} trading history {}'
                ''.format(
                    name,
                    ticker,
                    use_log))
    # if publish an trading history dataset

    if should_publish_report_dataset and dataset_publish_report:
        s3_log = ''
        redis_log = ''
        file_log = ''
        use_log = 'publish'

        if (report_config['redis_address'] and
                report_config['redis_db'] and
                report_config['redis_key']):
            redis_log = 'redis://{}@{}/{}'.format(
                report_config['redis_address'],
                report_config['redis_db'],
                report_config['redis_key'])
            use_log += ' {}'.format(
                redis_log)
        else:
            report_config['redis_enabled'] = False
        if (report_config['s3_address'] and
                report_config['s3_bucket'] and
                report_config['s3_key']):
            s3_log = 's3://{}/{}/{}'.format(
                report_config['s3_address'],
                report_config['s3_bucket'],
                report_config['s3_key'])
            use_log += ' {}'.format(
                s3_log)
        else:
            report_config['s3_enabled'] = False
        if report_config['output_file']:
            file_log = ' file:{}'.format(
                report_config['output_file'])
            use_log += ' {}'.format(
                file_log)

        if verbose:
            log.info(
                '{} - publishing ticker={} trading performance report {}'
                ''.format(
                    name,
                    ticker,
                    use_log))

        publish_status = algo.publish_report_dataset(
            **report_config)
        if publish_status != ae_consts.SUCCESS:
            msg = (
                'failed to publish trading performance report datasets '
                'with status {} attempted to {}'.format(
                    ae_consts.get_status(status=publish_status),
                    use_log))
            log.error(msg)
            return build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=None)

        if verbose:
            log.info(
                '{} - publish - done ticker={} trading performance report {}'
                ''.format(
                    name,
                    ticker,
                    use_log))
    # if publish an trading performance report dataset

    if verbose:
        log.info(
            '{} - done publishing datasets for ticker={} from {} to {}'.format(
                name,
                ticker,
                use_start_date,
                use_end_date))

    return algo_res
# end of run_custom_algo



"""
Build a single indicator for an algorithm
"""

import uuid
import copy
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def build_indicator_node(
        node,
        label=None):
    """build_indicator_node

    Parse a dictionary in the algorithm config ``indicators`` list
    and return a dicationry

    Supported values found in:
    `analysis_engine/consts.py <https://github.com/AlgoTraders/sto
    ck-analysis-engine/blob/master/analysis_engine/consts.py>`__

    :param node: single dictionary from the config's ``indicators`` list
    :param label: optional - string log tracking
        this class in the logs (usually just the algo
        name is good enough to help debug issues
        when running distributed)
    :return: dictionary
    """
    if not label:
        label = 'build_indicator_node'

    name = node.get(
        'name',
        None)
    if not name:
        raise Exception(
            '{} - missing "name" in indicator dictionary={}'.format(
                label,
                node))
    # end of name check

    ind_id = str(uuid.uuid4()).replace('-', '')
    uses_dataset_str = node.get(
        'uses_data',
        'daily')
    uses_dataset = ae_consts.get_indicator_uses_data_as_int(
        val=uses_dataset_str)
    if uses_dataset == ae_consts.INDICATOR_USES_DATA_UNSUPPORTED:
        uses_dataset = ae_consts.INDICATOR_USES_DATA_ANY
        log.debug(
            '{} - unsupported indicator '
            'uses_dataset={} defaulting to "daily"'.format(
                label,
                uses_dataset_str))
    # end of supported indicator dataset types

    ind_category_str = node.get(
        'category',
        'momentum')
    ind_category = ae_consts.get_indicator_category_as_int(
        val=ind_category_str)
    if ind_category == ae_consts.INDICATOR_CATEGORY_UNKNOWN:
        ind_category = ae_consts.INDICATOR_CATEGORY_MOMENTUM
        log.debug(
            '{} - unsupported indicator '
            'category={} defaulting to "momentum"'.format(
                label,
                ind_category))
    # end of supported indicator category

    ind_type_str = node.get(
        'type',
        'technical')
    ind_type = ae_consts.get_indicator_type_as_int(
        val=ind_type_str)
    if ind_type == ae_consts.INDICATOR_TYPE_UNKNOWN:
        ind_type = ae_consts.INDICATOR_TYPE_TECHNICAL
        log.debug(
            '{} - unsupported indicator '
            'type={} defaulting to "technical"'.format(
                label,
                ind_type))
    # end of supported indicator type

    # allow easier key discovery
    use_unique_id = node.get(
        'unique_id',
        False)
    ind_name = '{}_{}_{}_{}'.format(
        name,
        ind_category,
        ind_type,
        uses_dataset)
    if use_unique_id:
        ind_name = '{}_{}_{}_{}_{}'.format(
            name,
            ind_category,
            ind_type,
            uses_dataset,
            ind_id)

    use_module_name = None
    use_path_to_module = None

    # none will use the BaseIndicator which does nothing
    use_path_to_module = node.get(
        'module_path',
        ae_consts.INDICATOR_BASE_MODULE_PATH)
    if not use_path_to_module:
        raise Exception(
            'Failed building Indicator node with missing '
            'module_path node={}'.format(
                node))
    use_module_name = node.get(
        'module_name',
        node.get(
            'name',
            ind_id))

    default_report_ignore_keys = \
        ae_consts.INDICATOR_IGNORED_CONIGURABLE_KEYS

    report_dict = {
        'id': ind_id,
        'name': ind_name,
        'created': ae_utils.utc_now_str(),
        'version': 1,
        'module_name': use_module_name,
        'path_to_module': use_path_to_module,
        'report_ignore_keys': default_report_ignore_keys,
        'metrics': {
            'type': ind_type,
            'category': ind_category,
            'uses_data': uses_dataset
        }
    }

    labeled_node = copy.deepcopy(node)

    # allow a node's sub report dir to be patched with this
    # tracking + reporting data
    # the algorithms will flatten:
    #    indicators[ind_name]['report']['metrics']
    # for trading performance report generation
    if 'report' in labeled_node:
        labeled_node['report'].update(report_dict)
    else:
        labeled_node['report'] = report_dict

    return labeled_node
# end of build_indicator_node



#!/usr/bin/env python

import celery
import analysis_engine.work_tasks.get_celery_app as get_celery_app
import analysis_engine.consts as consts
import spylunking.log.setup_logging as log_utils


# Disable celery log hijacking
# https://github.com/celery/celery/issues/2509
@celery.signals.setup_logging.connect
def setup_celery_logging(**kwargs):
    pass


log = log_utils.build_colorized_logger(
    name=consts.APP_NAME,
    log_config_path=consts.LOG_CONFIG_PATH)

log.info(
    'start - {}'.format(
        consts.APP_NAME))

log.info(
    'broker={} backend={} '
    'config={} include_tasks={}'.format(
        consts.WORKER_BROKER_URL,
        consts.WORKER_BACKEND_URL,
        consts.WORKER_CELERY_CONFIG_MODULE,
        consts.WORKER_TASKS))

# Get the Celery app from the project's get_celery_app module
app = get_celery_app.get_celery_app(
    name=consts.APP_NAME,
    path_to_config_module=consts.WORKER_CELERY_CONFIG_MODULE,
    auth_url=consts.WORKER_BROKER_URL,
    backend_url=consts.WORKER_BACKEND_URL,
    include_tasks=consts.INCLUDE_TASKS)

log.info('starting celery')
app.start()

log.info(
    'end - {}'.format(
        consts.APP_NAME))



"""
Consts and helper functions

Algorithm Environment Variables
-------------------------------

::

    ALGO_MODULE_PATH = ev(
        'ALGO_MODULE_PATH',
        '/opt/sa/analysis_engine/mocks/example_algo_minute.py')
    ALGO_BASE_MODULE_PATH = ev(
        'ALGO_BASE_MODULE_PATH',
        '/opt/sa/analysis_engine/algo.py')
    ALGO_MODULE_NAME = ev(
        'ALGO_MODULE_NAME',
        'example_algo_minute')
    ALGO_VERSION = ev(
        'ALGO_VERSION',
        '1')
    ALGO_BUYS_S3_BUCKET_NAME = ev(
        'ALGO_BUYS_S3_BUCKET_NAME',
        'algobuys')
    ALGO_SELLS_S3_BUCKET_NAME = ev(
        'ALGO_SELLS_S3_BUCKET_NAME',
        'algosells')
    ALGO_RESULT_S3_BUCKET_NAME = ev(
        'ALGO_RESULT_S3_BUCKET_NAME',
        'algoresults')
    ALGO_READY_DATASET_S3_BUCKET_NAME = ev(
        'ALGO_READY_DATASET_S3_BUCKET_NAME',
        'algoready')
    ALGO_EXTRACT_DATASET_S3_BUCKET_NAME = ev(
        'ALGO_EXTRACT_DATASET_S3_BUCKET_NAME',
        'algoready')
    ALGO_HISTORY_DATASET_S3_BUCKET_NAME = ev(
        'ALGO_HISTORY_DATASET_S3_BUCKET_NAME',
        'algohistory')
    ALGO_REPORT_DATASET_S3_BUCKET_NAME = ev(
        'ALGO_REPORT_DATASET_S3_BUCKET_NAME',
        'algoreport')
    ALGO_BACKUP_DATASET_S3_BUCKET_NAME = ev(
        'ALGO_BACKUP_DATASET_S3_BUCKET_NAME',
        'algobackup')
    ALGO_READY_DIR = ev(
        'ALGO_READY_DIR',
        '/tmp')
    ALGO_EXTRACT_DIR = ev(
        'ALGO_EXTRACT_DIR',
        '/tmp')
    ALGO_HISTORY_DIR = ev(
        'ALGO_HISTORY_HISTORY_DIR',
        '/tmp')
    ALGO_REPORT_DIR = ev(
        'ALGO_REPORT_DIR',
        '/tmp')
    ALGO_LOAD_DIR = ev(
        'ALGO_LOAD_DIR',
        '/tmp')
    ALGO_BACKUP_DIR = ev(
        'ALGO_BACKUP_DIR',
        '/tmp')
    ALGO_READY_REDIS_ADDRESS = ev(
        'ALGO_READY_REDIS_ADDRESS',
        'localhost:6379')
    ALGO_EXTRACT_REDIS_ADDRESS = ev(
        'ALGO_EXTRACT_REDIS_ADDRESS',
        'localhost:6379')
    ALGO_HISTORY_REDIS_ADDRESS = ev(
        'ALGO_HISTORY_REDIS_ADDRESS',
        'localhost:6379')
    ALGO_REPORT_REDIS_ADDRESS = ev(
        'ALGO_REPORT_REDIS_ADDRESS',
        'localhost:6379')
    ALGO_BACKUP_REDIS_ADDRESS = ev(
        'ALGO_BACKUP_REDIS_ADDRESS',
        'localhost:6379')
    ALGO_HISTORY_VERSION = ev(
        'ALGO_HISTORY_VERSION',
        '1')
    ALGO_REPORT_VERSION = ev(
        'ALGO_REPORT_VERSION',
        '1')

Stock and Analysis Environment Variables
----------------------------------------

::

    TICKER = ev(
        'TICKER',
        'SPY')
    TICKER_ID = int(ev(
        'TICKER_ID',
        '1'))
    DEFAULT_TICKERS = ev(
        'DEFAULT_TICKERS',
        'SPY,AMZN,TSLA,NFLX').split(',')
    NEXT_EXP = analysis_engine.options_dates.option_expiration()
    NEXT_EXP_STR = NEXT_EXP.strftime('%Y-%m-%d')

Logging Environment Variables
-----------------------------

::

    APP_NAME = 'pr'
    LOG_CONFIG_PATH = ev(
        'LOG_CONFIG_PATH',
        './analysis_engine/log/logging.json')

Celery Environment Variables
----------------------------

::

    SLACK_WEBHOOK = ev(
        'SLACK_WEBHOOK',
        None)
    PROD_SLACK_ALERTS = ev(
        'PROD_SLACK_ALERTS',
        '0')
    SSL_OPTIONS = {}
    TRANSPORT_OPTIONS = {}
    WORKER_BROKER_URL = ev(
        'WORKER_BROKER_URL',
        'redis://localhost:6379/13')
    WORKER_BACKEND_URL = ev(
        'WORKER_BACKEND_URL',
        'redis://localhost:6379/14')
    WORKER_CELERY_CONFIG_MODULE = ev(
        'WORKER_CELERY_CONFIG_MODULE',
        'analysis_engine.work_tasks.celery_config')
    WORKER_TASKS = ev(
        'WORKER_TASKS',
        ('analysis_engine.work_tasks.get_new_pricing_data,'
         'analysis_engine.work_tasks.handle_pricing_update_task,'
         'analysis_engine.work_tasks.prepare_pricing_dataset,'
         'analysis_engine.work_tasks.publish_from_s3_to_redis,'
         'analysis_engine.work_tasks.publish_pricing_update,'
         'analysis_engine.work_tasks.task_screener_analysis,'
         'analysis_engine.work_tasks.publish_ticker_aggregate_from_s3'))
    INCLUDE_TASKS = WORKER_TASKS.split(',')

Supported S3 Environment Variables
----------------------------------

::

    ENABLED_S3_UPLOAD = ev(
        'ENABLED_S3_UPLOAD',
        '0') == '1'
    S3_ACCESS_KEY = ev(
        'AWS_ACCESS_KEY_ID',
        'trexaccesskey')
    S3_SECRET_KEY = ev(
        'AWS_SECRET_ACCESS_KEY',
        'trex123321')
    S3_REGION_NAME = ev(
        'AWS_DEFAULT_REGION',
        'us-east-1')
    S3_ADDRESS = ev(
        'S3_ADDRESS',
        'localhost:9000')
    S3_SECURE = ev(
        'S3_SECURE',
        '0') == '1'
    S3_BUCKET = ev(
        'S3_BUCKET',
        'pricing')
    S3_COMPILED_BUCKET = ev(
        'S3_COMPILED_BUCKET',
        'compileddatasets')
    S3_KEY = ev(
        'S3_KEY',
        'test_key')
    DAILY_S3_BUCKET_NAME = ev(
        'DAILY_S3_BUCKET_NAME',
        'daily')
    MINUTE_S3_BUCKET_NAME = ev(
        'MINUTE_S3_BUCKET_NAME',
        'minute')
    QUOTE_S3_BUCKET_NAME = ev(
        'QUOTE_S3_BUCKET_NAME',
        'quote')
    STATS_S3_BUCKET_NAME = ev(
        'STATS_S3_BUCKET_NAME',
        'stats')
    PEERS_S3_BUCKET_NAME = ev(
        'PEERS_S3_BUCKET_NAME',
        'peers')
    NEWS_S3_BUCKET_NAME = ev(
        'NEWS_S3_BUCKET_NAME',
        'news')
    FINANCIALS_S3_BUCKET_NAME = ev(
        'FINANCIALS_S3_BUCKET_NAME',
        'financials')
    EARNINGS_S3_BUCKET_NAME = ev(
        'EARNINGS_S3_BUCKET_NAME',
        'earnings')
    DIVIDENDS_S3_BUCKET_NAME = ev(
        'DIVIDENDS_S3_BUCKET_NAME',
        'dividends')
    COMPANY_S3_BUCKET_NAME = ev(
        'COMPANY_S3_BUCKET_NAME',
        'company')
    PREPARE_S3_BUCKET_NAME = ev(
        'PREPARE_S3_BUCKET_NAME',
        'prepared')
    ANALYZE_S3_BUCKET_NAME = ev(
        'ANALYZE_S3_BUCKET_NAME',
        'analyzed')
    SCREENER_S3_BUCKET_NAME = ev(
        'SCREENER_S3_BUCKET_NAME',
        'screener-data')
    PRICING_S3_BUCKET_NAME = ev(
        'PRICING_S3_BUCKET_NAME',
        'pricing')
    OPTIONS_S3_BUCKET_NAME = ev(
        'OPTIONS_S3_BUCKET_NAME',
        'options')

Supported Redis Environment Variables
-------------------------------------

::

    ENABLED_REDIS_PUBLISH = ev(
        'ENABLED_REDIS_PUBLISH',
        '0') == '1'
    REDIS_ADDRESS = ev(
        'REDIS_ADDRESS',
        'localhost:6379')
    REDIS_KEY = ev(
        'REDIS_KEY',
        'test_redis_key')
    REDIS_PASSWORD = ev(
        'REDIS_PASSWORD',
        None)
    REDIS_DB = int(ev(
        'REDIS_DB',
        '0'))
    REDIS_EXPIRE = ev(
        'REDIS_EXPIRE',
        None)

"""

import os
import sys
import json
import analysis_engine.options_dates


def ev(
        k,
        v):
    '''ev

    :param k: environment variable key
    :param v: environment variable value
    '''
    val = os.getenv(k, v)
    if val:
        return val.strip()
    return val
# end of ev


SUCCESS = 0
FAILED = 1
ERR = 2
EX = 3
NOT_RUN = 4
INVALID = 5
NOT_DONE = 6
NOT_SET = 7
EMPTY = 8
TRADE_OPEN = 9
TRADE_NOT_ENOUGH_FUNDS = 10
TRADE_FILLED = 11
TRADE_NO_SHARES_TO_SELL = 12
TRADE_EXPIRED = 13
TRADE_SHARES = 14
TRADE_VERTICAL_BULL_SPREAD = 15
TRADE_VERTICAL_BEAR_SPREAD = 16
TRADE_PROFITABLE = 17
TRADE_NOT_PROFITABLE = 18
TRADE_HIT_STOP_LOSS = 19
TRADE_HIT_STOP_LOSS_PERCENT = 20
TRADE_HIT_TAILING_STOP_LOSS = 21
TRADE_HIT_TAILING_STOP_LOSS_PERCENT = 22
TRADE_INVALID = 23
TRADE_ERROR = 24
TRADE_ENTRY = 25
TRADE_EXIT = 26
BACKTEST_FOUND_TRADE_PROFITABLE = 27
BACKTEST_FOUND_TRADE_NOT_PROFITABLE = 28
BACKTEST_FOUND_TRADE_NEVER_FILLED = 29  # limit order price never hit
BACKTEST_FOUND_TRADE_EXPIRED = 30  # trades assumed are expired after a day
SPREAD_VERTICAL_BULL = 31
SPREAD_VERTICAL_BEAR = 32
OPTION_CALL = 33
OPTION_PUT = 34
ALGO_PROFITABLE = 35
ALGO_NOT_PROFITABLE = 36
ALGO_ERROR = 37
ALGO_NOT_ACTIVE = 38
S3_FAILED = 39
REDIS_FAILED = 40
FILE_FAILED = 41
SLACK_FAILED = 42
ALGO_TIMESERIES_DAY = 43  # evaluate trade performance on daily-units
ALGO_TIMESERIES_MINUTE = 44  # evaluate trade performance on minute-units
ALGO_TRADE_INDICATOR_COUNTS = 45  # trade off num indicators said buy/sell

INDICATOR_CATEGORY_MOMENTUM = 60
INDICATOR_CATEGORY_OVERLAP = 61
INDICATOR_CATEGORY_PRICE = 62
INDICATOR_CATEGORY_VOLUME = 63
INDICATOR_CATEGORY_VOLATILITY = 64
INDICATOR_CATEGORY_SINGLE_CALL = 65
INDICATOR_CATEGORY_SINGLE_PUT = 66
INDICATOR_CATEGORY_BULL_CALL = 67
INDICATOR_CATEGORY_BEAR_PUT = 68
INDICATOR_CATEGORY_QUARTERLY = 69
INDICATOR_CATEGORY_YEARLY = 70
INDICATOR_CATEGORY_INCOME_STMT = 71
INDICATOR_CATEGORY_CASH_FLOW = 72
INDICATOR_CATEGORY_BALANCE_SHEET = 73
INDICATOR_CATEGORY_PRESS_RELEASE = 74
INDICATOR_CATEGORY_CUSTOM = 75
INDICATOR_CATEGORY_NEWS = 76
INDICATOR_CATEGORY_EARNINGS = 77
INDICATOR_CATEGORY_CSUITE = 78
INDICATOR_CATEGORY_SPLITS = 79
INDICATOR_CATEGORY_REVERSE_SPLITS = 80
INDICATOR_CATEGORY_DISTRIBUTIONS = 81
INDICATOR_CATEGORY_SPINOFFS = 82
INDICATOR_CATEGORY_MERGER_ACQ = 83
INDICATOR_CATEGORY_EXCHANGE_INCLUSION = 84
INDICATOR_CATEGORY_EXCHANGE_EXCLUSION = 85
INDICATOR_CATEGORY_TRIAL_POSITIVE = 86
INDICATOR_CATEGORY_TRIAL_NEGATIVE = 87
INDICATOR_CATEGORY_SHORT_SELLERS = 88
INDICATOR_CATEGORY_REAL_ESTATE = 89
INDICATOR_CATEGORY_HOUSING = 90
INDICATOR_CATEGORY_PIPELINE = 91
INDICATOR_CATEGORY_CONSTRUCTION = 94
INDICATOR_CATEGORY_FED = 93
INDICATOR_CATEGORY_UNKNOWN = 94

INDICATOR_TYPE_TECHNICAL = 200
INDICATOR_TYPE_FUNDAMENTAL = 201
INDICATOR_TYPE_NEWS = 202
INDICATOR_TYPE_SECTOR = 203
INDICATOR_TYPE_MARKET = 204
INDICATOR_TYPE_DIVIDEND = 205
INDICATOR_TYPE_CUSTOM = 206
INDICATOR_TYPE_UNKNOWN = 207

INDICATOR_USES_DAILY_DATA = 500
INDICATOR_USES_MINUTE_DATA = 501
INDICATOR_USES_QUOTE_DATA = 502
INDICATOR_USES_STATS_DATA = 503
INDICATOR_USES_PEERS_DATA = 504
INDICATOR_USES_NEWS_DATA = 505
INDICATOR_USES_FINANCIAL_DATA = 506
INDICATOR_USES_EARNINGS_DATA = 507
INDICATOR_USES_DIVIDENDS_DATA = 508
INDICATOR_USES_COMPANY_DATA = 509
INDICATOR_USES_PRICING_DATA = 510
INDICATOR_USES_OPTIONS_DATA = 511
INDICATOR_USES_CALLS_DATA = 512
INDICATOR_USES_PUTS_DATA = 513
INDICATOR_USES_DATA_UNSUPPORTED = 514
INDICATOR_USES_DATA_ANY = 515

INT_INDICATOR_NOT_PROCESSED = 600
INT_INDICATOR_IGNORE_ACTION = 601
INT_INDICATOR_BUY_ACTION = 602
INT_INDICATOR_SELL_ACTION = 603

INDICATOR_BUY = 'buy'
INDICATOR_SELL = 'sell'
INDICATOR_IGNORE = 'ignore'
INDICATOR_RESET = 'reset'

SA_MODE_PREPARE = 10000
SA_MODE_ANALYZE = 10001
SA_MODE_PREDICT = 10002
SA_MODE_EXTRACT = 10003
SA_MODE_SHOW_DATASET = 10004
SA_MODE_RESTORE_REDIS_DATASET = 10005
SA_MODE_RUN_ALGO = 10006
SA_MODE_SHOW_HISTORY_DATASET = 10007
SA_MODE_SHOW_REPORT_DATASET = 10008

SA_DATASET_TYPE_ALGO_READY = 20000
SA_DATASET_TYPE_TRADING_HISTORY = 20001
SA_DATASET_TYPE_TRADING_REPORT = 20002

PLOT_ACTION_SHOW = 20000
PLOT_ACTION_SAVE_TO_S3 = 20001
PLOT_ACTION_SAVE_AS_FILE = 20002

FETCH_MODE_ALL = 30000
FETCH_MODE_YHO = 30001
FETCH_MODE_IEX = 30002

# version of python
IS_PY2 = sys.version[0] == '2'
NUM_BYTES_IN_AN_MB = 1048576

APP_NAME = ev(
    'APP_NAME',
    'pr')
LOG_CONFIG_PATH = ev(
    'LOG_CONFIG_PATH',
    './analysis_engine/log/logging.json')
SSL_OPTIONS = {}
TRANSPORT_OPTIONS = {}
WORKER_BROKER_URL = ev(
    'WORKER_BROKER_URL',
    'redis://localhost:6379/13')
WORKER_BACKEND_URL = ev(
    'WORKER_BACKEND_URL',
    'redis://localhost:6379/14')
WORKER_CELERY_CONFIG_MODULE = ev(
    'WORKER_CELERY_CONFIG_MODULE',
    'analysis_engine.work_tasks.celery_config')
WORKER_TASKS = ev(
    'WORKER_TASKS',
    ('analysis_engine.work_tasks.get_new_pricing_data,'
     'analysis_engine.work_tasks.handle_pricing_update_task,'
     'analysis_engine.work_tasks.prepare_pricing_dataset,'
     'analysis_engine.work_tasks.publish_from_s3_to_redis,'
     'analysis_engine.work_tasks.publish_pricing_update,'
     'analysis_engine.work_tasks.task_screener_analysis,'
     'analysis_engine.work_tasks.run_distributed_algorithm,'
     'analysis_engine.work_tasks.publish_ticker_aggregate_from_s3'
     ''))
INCLUDE_TASKS = WORKER_TASKS.split(',')
CELERY_DISABLED = ev('CELERY_DISABLED', '0') == '1'

########################################
#
# Custom Variables
#
########################################
TICKER = ev(
    'TICKER',
    'SPY')
TICKER_ID = int(ev(
    'TICKER_ID',
    '1'))
DEFAULT_TICKERS = ev(
    'DEFAULT_TICKERS',
    'SPY,AMZN,TSLA,NFLX').split(',')
NEXT_EXP = analysis_engine.options_dates.option_expiration()
NEXT_EXP_STR = NEXT_EXP.strftime('%Y-%m-%d')
DAILY_S3_BUCKET_NAME = ev(
    'DAILY_S3_BUCKET_NAME',
    'daily')
MINUTE_S3_BUCKET_NAME = ev(
    'MINUTE_S3_BUCKET_NAME',
    'minute')
QUOTE_S3_BUCKET_NAME = ev(
    'QUOTE_S3_BUCKET_NAME',
    'quote')
STATS_S3_BUCKET_NAME = ev(
    'STATS_S3_BUCKET_NAME',
    'stats')
PEERS_S3_BUCKET_NAME = ev(
    'PEERS_S3_BUCKET_NAME',
    'peers')
NEWS_S3_BUCKET_NAME = ev(
    'NEWS_S3_BUCKET_NAME',
    'news')
FINANCIALS_S3_BUCKET_NAME = ev(
    'FINANCIALS_S3_BUCKET_NAME',
    'financials')
EARNINGS_S3_BUCKET_NAME = ev(
    'EARNINGS_S3_BUCKET_NAME',
    'earnings')
DIVIDENDS_S3_BUCKET_NAME = ev(
    'DIVIDENDS_S3_BUCKET_NAME',
    'dividends')
COMPANY_S3_BUCKET_NAME = ev(
    'COMPANY_S3_BUCKET_NAME',
    'company')
FETCH_MODE = ev(
    'FETCH_MODE',
    'full')
PREPARE_S3_BUCKET_NAME = ev(
    'PREPARE_S3_BUCKET_NAME',
    'prepared')
ANALYZE_S3_BUCKET_NAME = ev(
    'ANALYZE_S3_BUCKET_NAME',
    'analyzed')
SCREENER_S3_BUCKET_NAME = ev(
    'SCREENER_S3_BUCKET_NAME',
    'screener-data')
PRICING_S3_BUCKET_NAME = ev(
    'PRICING_S3_BUCKET_NAME',
    'pricing')
OPTIONS_S3_BUCKET_NAME = ev(
    'OPTIONS_S3_BUCKET_NAME',
    'options')
PREPARE_DATA_MIN_SIZE = 11
PLOT_COLORS = {
    'red': '#E74C3C',
    'feldspar': '#D19275',
    'copper': '#EDC393',
    'brown': '#6B4226',
    'orange': '#FF7D40',
    'maroon': '#800000',
    'gray': '#8B8989',
    'black': '#111111',
    'pink': '#FFCCCC',
    'green': '#2ECC71',
    'blue': '#3498db',
    'darkblue': '#000080',
    'lightgreen': '#C0FF3E',
    'darkgreen': '#385E0F',
    'gold': '#FFCC11',
    'yellow': '#FFE600',
    'volumetop': '#385E0F',
    'volume': '#ADFF2F',
    'high': '#CC1100',
    'low': '#164E71',
    'open': '#608DC0',
    'close': '#99CC32',
    'white': '#FFFFFF'
}

IEX_DAILY_DATE_FORMAT = '%Y-%b-%d'
IEX_MINUTE_DATE_FORMAT = '%Y-%m-%d %I:%M:%S %p'
IEX_TICK_DATE_FORMAT = '%Y-%m-%d %I:%M:%S %p'
IEX_QUOTE_DATE_FORMAT = '%B %d, %Y'
IEX_DATASETS_DEFAULT = [
    'daily',
    'minute',
    'quote',
    'stats',
    'peers',
    'news',
    'financials',
    'earnings',
    'dividends',
    'company'
]

COMMON_DATE_FORMAT = '%Y-%m-%d'
COMMON_TICK_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
CACHE_DICT_VERSION = 1

SLACK_WEBHOOK = ev(
    'SLACK_WEBHOOK',
    None)
PROD_SLACK_ALERTS = ev(
    'PROD_SLACK_ALERTS',
    '0')
DATASET_COLLECTION_VERSION = 1
DATASET_COLLECTION_SLACK_ALERTS = ev(
    'DATASET_COLLECTION_SLACK_ALERTS',
    '0')
SLACK_FINVIZ_ALL_COLUMNS = [
    'ticker',
    'price',
    'volume',
    'change',
    'pe',
    'market_cap',
    'company',
    'industry',
    'sector',
    'country'
]
SLACK_FINVIZ_COLUMNS = [
    'ticker',
    'price',
    'volume',
    'change',
    'pe',
    'market_cap'
]
ALGO_INPUT_COMPRESS = (ev(
    'ALGO_INPUT_COMPRESS',
    '0') == '1')
ALGO_LOAD_COMPRESS = (ev(
    'ALGO_LOAD_COMPRESS',
    '0') == '1')
ALGO_HISTORY_COMPRESS = (ev(
    'ALGO_HISTORY_COMPRESS',
    '0') == '1')
ALGO_HISTORY_VERSION = ev(
    'ALGO_HISTORY_VERSION',
    '1')
ALGO_REPORT_COMPRESS = (ev(
    'ALGO_REPORT_COMPRESS',
    '0') == '1')
ALGO_REPORT_VERSION = ev(
    'ALGO_REPORT_VERSION',
    '1')
DEFAULT_SERIALIZED_DATASETS = [
    'daily',
    'minute',
    'quote',
    'stats',
    'peers',
    'news1',
    'financials',
    'earnings',
    'dividends',
    'company',
    'news',
    'calls',
    'puts',
    'pricing'
]
EMPTY_DF_STR = '[{}]'
EMPTY_DF_LIST = [{}]

########################################
#
# Algorithm Variables
#
########################################
ALGO_MODULE_PATH = ev(
    'ALGO_MODULE_PATH',
    '/opt/sa/analysis_engine/mocks/example_algo_minute.py')
ALGO_BASE_MODULE_PATH = ev(
    'ALGO_BASE_MODULE_PATH',
    '/opt/sa/analysis_engine/algo.py')
ALGO_MODULE_NAME = ev(
    'ALGO_MODULE_NAME',
    'example_algo_minute')
ALGO_VERSION = ev(
    'ALGO_VERSION',
    '1')
ALGO_BUYS_S3_BUCKET_NAME = ev(
    'ALGO_BUYS_S3_BUCKET_NAME',
    'algobuys')
ALGO_SELLS_S3_BUCKET_NAME = ev(
    'ALGO_SELLS_S3_BUCKET_NAME',
    'algosells')
ALGO_RESULT_S3_BUCKET_NAME = ev(
    'ALGO_RESULT_S3_BUCKET_NAME',
    'algoresult')
ALGO_EXTRACT_DATASET_S3_BUCKET_NAME = ev(
    'ALGO_EXTRACT_DATASET_S3_BUCKET_NAME',
    'algoready')
ALGO_READY_DATASET_S3_BUCKET_NAME = ev(
    'ALGO_READY_DATASET_S3_BUCKET_NAME',
    'algoready')
ALGO_HISTORY_DATASET_S3_BUCKET_NAME = ev(
    'ALGO_HISTORY_DATASET_S3_BUCKET_NAME',
    'algohistory')
ALGO_REPORT_DATASET_S3_BUCKET_NAME = ev(
    'ALGO_REPORT_DATASET_S3_BUCKET_NAME',
    'algoreport')
ALGO_BACKUP_DATASET_S3_BUCKET_NAME = ev(
    'ALGO_BACKUP_DATASET_S3_BUCKET_NAME',
    'algobackup')
ALGO_READY_DIR = ev(
    'ALGO_READY_DIR',
    '/tmp')
ALGO_EXTRACT_DIR = ev(
    'ALGO_EXTRACT_DIR',
    '/tmp')
ALGO_HISTORY_DIR = ev(
    'ALGO_HISTORY_HISTORY_DIR',
    '/tmp')
ALGO_REPORT_DIR = ev(
    'ALGO_REPORT_DIR',
    '/tmp')
ALGO_LOAD_DIR = ev(
    'ALGO_LOAD_DIR',
    '/tmp')
ALGO_BACKUP_DIR = ev(
    'ALGO_BACKUP_DIR',
    '/tmp')
ALGO_READY_REDIS_ADDRESS = ev(
    'ALGO_READY_REDIS_ADDRESS',
    'localhost:6379')
ALGO_EXTRACT_REDIS_ADDRESS = ev(
    'ALGO_EXTRACT_REDIS_ADDRESS',
    'localhost:6379')
ALGO_HISTORY_REDIS_ADDRESS = ev(
    'ALGO_HISTORY_REDIS_ADDRESS',
    'localhost:6379')
ALGO_REPORT_REDIS_ADDRESS = ev(
    'ALGO_REPORT_REDIS_ADDRESS',
    'localhost:6379')
ALGO_BACKUP_REDIS_ADDRESS = ev(
    'ALGO_BACKUP_REDIS_ADDRESS',
    'localhost:6379')

########################################
#
# Indicator Variables
#
########################################
INDICATOR_BASE_MODULE = ev(
    'INDICATOR_BASE_MODULE',
    'analysis_engine.indicators.base_indicator.BaseIndicator')
INDICATOR_BASE_MODULE_PATH = ev(
    'INDICATOR_BASE_MODULE_PATH',
    'analysis_engine/indicators/base_indicator.py')
INDICATOR_IGNORED_CONIGURABLE_KEYS = [
    'name',
    'module_path',
    'category',
    'type',
    'uses_data',
    'report'
]

########################################
#
# S3 Variables
#
########################################
ENABLED_S3_UPLOAD = ev(
    'ENABLED_S3_UPLOAD',
    '0') == '1'
S3_ACCESS_KEY = ev(
    'AWS_ACCESS_KEY_ID',
    'trexaccesskey')
S3_SECRET_KEY = ev(
    'AWS_SECRET_ACCESS_KEY',
    'trex123321')
S3_REGION_NAME = ev(
    'AWS_DEFAULT_REGION',
    'us-east-1')
S3_ADDRESS = ev(
    'S3_ADDRESS',
    'localhost:9000')
S3_SECURE = ev(
    'S3_SECURE',
    '0') == '1'
S3_BUCKET = ev(
    'S3_BUCKET',
    'pricing')
S3_COMPILED_BUCKET = ev(
    'S3_COMPILED_BUCKET',
    'compileddatasets')
S3_KEY = ev(
    'S3_KEY',
    'test_key')

########################################
#
# Redis Variables
#
########################################
ENABLED_REDIS_PUBLISH = ev(
    'ENABLED_REDIS_PUBLISH',
    '0') == '1'
REDIS_ADDRESS = ev(
    'REDIS_ADDRESS',
    'localhost:6379')
REDIS_KEY = ev(
    'REDIS_KEY',
    'test_redis_key')
REDIS_PASSWORD = ev(
    'REDIS_PASSWORD',
    None)
REDIS_DB = int(ev(
    'REDIS_DB',
    '0'))
REDIS_EXPIRE = ev(
    'REDIS_EXPIRE',
    None)

# copy these values over
# when calling child tasks from a
# parent where the engine is
# running inside a fully-dockerized
# environment like kubernetes
# or docker-compose
SERVICE_VALS = [
    'ticker',
    's3_address',
    's3_access_key',
    's3_secret_key',
    's3_bucket',
    's3_secure',
    's3_region_name',
    'redis_address',
    'redis_db',
    'redis_password',
    'redis_expire'
]


def get_status(
        status):
    """get_status

    Return the string label for an integer status code
    which should be one of the ones above.

    :param status: integer status code
    """
    if status == SUCCESS:
        return 'SUCCESS'
    elif status == FAILED:
        return 'FAILED'
    elif status == ERR:
        return 'ERR'
    elif status == EX:
        return 'EX'
    elif status == NOT_RUN:
        return 'NOT_RUN'
    elif status == INVALID:
        return 'INVALID'
    elif status == NOT_DONE:
        return 'NOT_DONE'
    elif status == NOT_SET:
        return 'NOT_SET'
    elif status == EMPTY:
        return 'EMPTY'
    elif status == SA_MODE_PREPARE:
        return 'SA_MODE_PREPARE'
    elif status == SA_MODE_ANALYZE:
        return 'SA_MODE_ANALYZE'
    elif status == SA_MODE_PREDICT:
        return 'SA_MODE_PREDICT'
    elif status == SA_MODE_EXTRACT:
        return 'SA_MODE_EXTRACT'
    elif status == SA_MODE_SHOW_DATASET:
        return 'SA_MODE_SHOW_DATASET'
    elif status == SA_MODE_RESTORE_REDIS_DATASET:
        return 'SA_MODE_RESTORE_REDIS_DATASET'
    elif status == SA_MODE_RUN_ALGO:
        return 'SA_MODE_RUN_ALGO'
    elif status == SA_MODE_SHOW_HISTORY_DATASET:
        return 'SA_MODE_SHOW_HISTORY_DATASET'
    elif status == SA_MODE_SHOW_REPORT_DATASET:
        return 'SA_MODE_SHOW_REPORT_DATASET'
    elif status == PLOT_ACTION_SHOW:
        return 'PLOT_ACTION_SHOW'
    elif status == PLOT_ACTION_SAVE_TO_S3:
        return 'PLOT_ACTION_SAVE_TO_S3'
    elif status == PLOT_ACTION_SAVE_AS_FILE:
        return 'PLOT_ACTION_SAVE_AS_FILE'
    elif status == TRADE_OPEN:
        return 'TRADE_OPEN'
    elif status == TRADE_NOT_ENOUGH_FUNDS:
        return 'TRADE_NOT_ENOUGH_FUNDS'
    elif status == TRADE_FILLED:
        return 'TRADE_FILLED'
    elif status == TRADE_NO_SHARES_TO_SELL:
        return 'TRADE_NO_SHARES_TO_SELL'
    elif status == TRADE_EXPIRED:
        return 'TRADE_EXPIRED'
    elif status == TRADE_SHARES:
        return 'TRADE_SHARES'
    elif status == TRADE_VERTICAL_BULL_SPREAD:
        return 'TRADE_VERTICAL_BULL_SPREAD'
    elif status == TRADE_VERTICAL_BEAR_SPREAD:
        return 'TRADE_VERTICAL_BEAR_SPREAD'
    elif status == TRADE_PROFITABLE:
        return 'TRADE_PROFITABLE'
    elif status == TRADE_NOT_PROFITABLE:
        return 'TRADE_NOT_PROFITABLE'
    elif status == TRADE_HIT_STOP_LOSS:
        return 'TRADE_HIT_STOP_LOSS'
    elif status == TRADE_HIT_STOP_LOSS_PERCENT:
        return 'TRADE_HIT_STOP_LOSS_PERCENT'
    elif status == TRADE_HIT_TAILING_STOP_LOSS:
        return 'TRADE_HIT_TAILING_STOP_LOSS'
    elif status == TRADE_HIT_TAILING_STOP_LOSS_PERCENT:
        return 'TRADE_HIT_TAILING_STOP_LOSS_PERCENT'
    elif status == TRADE_INVALID:
        return 'TRADE_INVALID'
    elif status == TRADE_ERROR:
        return 'TRADE_ERROR'
    elif status == TRADE_ENTRY:
        return 'TRADE_ENTRY'
    elif status == TRADE_EXIT:
        return 'TRADE_EXIT'
    elif status == BACKTEST_FOUND_TRADE_PROFITABLE:
        return 'BACKTEST_FOUND_TRADE_PROFITABLE'
    elif status == BACKTEST_FOUND_TRADE_NOT_PROFITABLE:
        return 'BACKTEST_FOUND_TRADE_NOT_PROFITABLE'
    elif status == BACKTEST_FOUND_TRADE_NEVER_FILLED:
        return 'BACKTEST_FOUND_TRADE_NEVER_FILLED'
    elif status == BACKTEST_FOUND_TRADE_EXPIRED:
        return 'BACKTEST_FOUND_TRADE_EXPIRED'
    elif status == SPREAD_VERTICAL_BULL:
        return 'SPREAD_VERTICAL_BULL'
    elif status == SPREAD_VERTICAL_BEAR:
        return 'SPREAD_VERTICAL_BEAR'
    elif status == OPTION_CALL:
        return 'OPTION_CALL'
    elif status == OPTION_PUT:
        return 'OPTION_PUT'
    elif status == ALGO_PROFITABLE:
        return 'ALGO_PROFITABLE'
    elif status == ALGO_NOT_PROFITABLE:
        return 'ALGO_NOT_PROFITABLE'
    elif status == ALGO_ERROR:
        return 'ALGO_ERROR'
    elif status == ALGO_NOT_ACTIVE:
        return 'ALGO_NOT_ACTIVE'
    elif status == S3_FAILED:
        return 'S3_FAILED'
    elif status == REDIS_FAILED:
        return 'REDIS_FAILED'
    elif status == FILE_FAILED:
        return 'FILE_FAILED'
    elif status == SLACK_FAILED:
        return 'SLACK_FAILED'
    elif status == ALGO_TIMESERIES_DAY:
        return 'ALGO_TIMESERIES_DAY'
    elif status == ALGO_TIMESERIES_MINUTE:
        return 'ALGO_TIMESERIES_MINUTE'
    elif status == ALGO_TRADE_INDICATOR_COUNTS:
        return 'ALGO_TRADE_INDICATOR_COUNTS'
    elif status == SA_DATASET_TYPE_ALGO_READY:
        return 'SA_DATASET_TYPE_ALGO_READY'
    elif status == SA_DATASET_TYPE_TRADING_HISTORY:
        return 'SA_DATASET_TYPE_TRADING_HISTORY'
    elif status == SA_DATASET_TYPE_TRADING_REPORT:
        return 'SA_DATASET_TYPE_TRADING_REPORT'
    else:
        return 'unsupported status={}'.format(
            status)
# end of get_status


def ppj(
        json_data):
    """ppj

    :param json_data: dictionary to convert to
                      a pretty-printed, multi-line string
    """
    return str(
        json.dumps(
            json_data,
            sort_keys=True,
            indent=4,
            separators=(',', ': ')))
# end of ppj


def to_float_str(
        val):
    """to_float_str

    convert the float to a string with 2 decimal points of
    precision

    :param val: float to change to a 2-decimal string
    """
    return str("%0.2f" % float(val))
# end of to_float_str


def to_f(
        val):
    """to_f

    truncate the float to 2 decimal points of
    precision

    :param val: float to change
    """
    if val is None:
        return None

    return float(to_float_str(val))
# end of to_f


def get_mb(
        num):
    """get_mb

    convert a the number of bytes (as an ``integer``)
    to megabytes with 2 decimal points of precision

    :param num: integer - number of bytes
    """
    return to_f(num / NUM_BYTES_IN_AN_MB)
# end get_mb


def get_percent_done(
        progress,
        total):
    """get_percent_done

    calculate percentage done to 2 decimal points of
    precision

    :param progress: progress counter
    :param total: total number of counts
    """
    return to_f(float(float(progress)/float(total)*100.00))
# end of get_percent_done


def is_celery_disabled(
        work_dict=None):
    """is_celery_disabled

    :param work_dict: request to check
    """
    env_disabled = ev('CELERY_DISABLED', '0') == '1'
    request_disabled = False
    if work_dict:
        request_disabled = work_dict.get('celery_disabled', False)
    return (env_disabled or request_disabled)
# end of is_celery_disabled


INDICATOR_TYPE_MAPPING = {
    'technical': INDICATOR_TYPE_TECHNICAL,
    'fundamental': INDICATOR_TYPE_FUNDAMENTAL,
    'news': INDICATOR_TYPE_NEWS,
    'sector': INDICATOR_TYPE_SECTOR,
    'market': INDICATOR_TYPE_MARKET,
    'dividend': INDICATOR_TYPE_DIVIDEND,
    'custom': INDICATOR_TYPE_CUSTOM,
    'unknown': INDICATOR_TYPE_UNKNOWN
}


INDICATOR_CATEGORY_MAPPING = {
    'momentum': INDICATOR_CATEGORY_MOMENTUM,
    'overlap': INDICATOR_CATEGORY_OVERLAP,
    'price': INDICATOR_CATEGORY_PRICE,
    'volume': INDICATOR_CATEGORY_VOLUME,
    'volatility': INDICATOR_CATEGORY_VOLATILITY,
    'single_call': INDICATOR_CATEGORY_SINGLE_CALL,
    'single_put': INDICATOR_CATEGORY_SINGLE_PUT,
    'bull_call': INDICATOR_CATEGORY_BULL_CALL,
    'bear_put': INDICATOR_CATEGORY_BEAR_PUT,
    'quarterly': INDICATOR_CATEGORY_QUARTERLY,
    'yearly': INDICATOR_CATEGORY_YEARLY,
    'income_statement': INDICATOR_CATEGORY_INCOME_STMT,
    'cash_flow': INDICATOR_CATEGORY_CASH_FLOW,
    'balance_sheet': INDICATOR_CATEGORY_BALANCE_SHEET,
    'press_release': INDICATOR_CATEGORY_PRESS_RELEASE,
    'custom': INDICATOR_CATEGORY_CUSTOM,
    'news': INDICATOR_CATEGORY_NEWS,
    'earnings': INDICATOR_CATEGORY_EARNINGS,
    'csuite': INDICATOR_CATEGORY_CSUITE,
    'splits': INDICATOR_CATEGORY_SPLITS,
    'rev_splits': INDICATOR_CATEGORY_REVERSE_SPLITS,
    'distributions': INDICATOR_CATEGORY_DISTRIBUTIONS,
    'spinoffs': INDICATOR_CATEGORY_SPINOFFS,
    'merger_acq': INDICATOR_CATEGORY_MERGER_ACQ,
    'exchange_inclusion': INDICATOR_CATEGORY_EXCHANGE_INCLUSION,
    'exchange_exclusion': INDICATOR_CATEGORY_EXCHANGE_EXCLUSION,
    'trial_positive': INDICATOR_CATEGORY_TRIAL_POSITIVE,
    'trial_negative': INDICATOR_CATEGORY_TRIAL_NEGATIVE,
    'short_sellers': INDICATOR_CATEGORY_SHORT_SELLERS,
    'real_estate': INDICATOR_CATEGORY_REAL_ESTATE,
    'housing': INDICATOR_CATEGORY_HOUSING,
    'pipeline': INDICATOR_CATEGORY_PIPELINE,
    'construction': INDICATOR_CATEGORY_CONSTRUCTION,
    'fed': INDICATOR_CATEGORY_FED,
    'unknown': INDICATOR_CATEGORY_UNKNOWN,
}

INDICATOR_USES_DATA_MAPPING = {
    'daily': INDICATOR_USES_DAILY_DATA,
    'minute': INDICATOR_USES_MINUTE_DATA,
    'quote': INDICATOR_USES_QUOTE_DATA,
    'stats': INDICATOR_USES_STATS_DATA,
    'peers': INDICATOR_USES_PEERS_DATA,
    'news': INDICATOR_USES_NEWS_DATA,
    'financials': INDICATOR_USES_FINANCIAL_DATA,
    'earnings': INDICATOR_USES_EARNINGS_DATA,
    'dividends': INDICATOR_USES_DIVIDENDS_DATA,
    'company': INDICATOR_USES_COMPANY_DATA,
    'pricing': INDICATOR_USES_PRICING_DATA,
    'options': INDICATOR_USES_OPTIONS_DATA,
    'calls': INDICATOR_USES_CALLS_DATA,
    'puts': INDICATOR_USES_PUTS_DATA,
    'unsupported': INDICATOR_USES_DATA_UNSUPPORTED,
    'any': INDICATOR_USES_DATA_ANY
}


INDICATOR_ACTIONS = {
    INDICATOR_RESET: INT_INDICATOR_NOT_PROCESSED,
    INDICATOR_IGNORE: INT_INDICATOR_IGNORE_ACTION,
    INDICATOR_BUY: INT_INDICATOR_BUY_ACTION,
    INDICATOR_SELL: INT_INDICATOR_SELL_ACTION
}


ALGO_TIMESERIES = {
    'daily': ALGO_TIMESERIES_DAY,
    'minute': ALGO_TIMESERIES_MINUTE,
    'intraday': ALGO_TIMESERIES_MINUTE
}


def get_indicator_type_as_int(
        val=None):
    """get_indicator_type_as_int

    convert the string to the ``INDICATOR_TYPE_MAPPING``
    integer value

    :param val: integer to lookup in the ``INDICATOR_TYPE_MAPPING``
        dictionary
    """
    if not val:
        return INDICATOR_TYPE_UNKNOWN
    else:
        if val in INDICATOR_TYPE_MAPPING:
            return INDICATOR_TYPE_MAPPING[val]
        else:
            return INDICATOR_TYPE_UNKNOWN
    # if supported or not
# end of get_indicator_type_as_int


def get_indicator_category_as_int(
        val=None):
    """get_indicator_category_as_int

    convert the string to the ``INDICATOR_CATEGORY_MAPPING``
    integer value

    :param val: integer to lookup in the ``INDICATOR_CATEGORY_MAPPING``
        dictionary
    """
    if not val:
        return INDICATOR_CATEGORY_UNKNOWN
    else:
        if val in INDICATOR_CATEGORY_MAPPING:
            return INDICATOR_CATEGORY_MAPPING[val]
        else:
            return INDICATOR_CATEGORY_UNKNOWN
    # if supported or not
# end of get_indicator_category_as_int


def get_indicator_uses_data_as_int(
        val=None):
    """get_indicator_uses_data_as_int

    convert the string to the ``INDICATOR_USES_DATA_MAPPING``
    integer value

    :param val: integer to lookup in the ``INDICATOR_USES_DATA_MAPPING``
        dictionary
    """
    if not val:
        return INDICATOR_USES_DATA_ANY
    else:
        if val in INDICATOR_USES_DATA_MAPPING:
            return INDICATOR_USES_DATA_MAPPING[val]
        else:
            return INDICATOR_USES_DATA_UNSUPPORTED
    # if supported or not
# end of get_indicator_uses_data_as_int


def get_algo_timeseries_from_int(
        val):
    """get_algo_timeseries_from_int

    convert the integer value to the timeseries string
    found in the ``analysis_engine.consts.ALGO_TIMESERIES``
    dictionary

    :param val: integer value for finding the
        string timeseries label
    """
    for a in ALGO_TIMESERIES:
        if ALGO_TIMESERIES[a] == val:
            return a
    return (
        'unsupported algorithm timeseries value={}'.format(
            val))
# end of get_algo_timeseries_from_int



"""
Build a dictionary for running an algorithm
"""

import datetime
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def build_algo_request(
        ticker=None,
        tickers=None,
        use_key=None,
        start_date=None,
        end_date=None,
        datasets=None,
        balance=None,
        commission=None,
        num_shares=None,
        config_file=None,
        config_dict=None,
        load_config=None,
        history_config=None,
        report_config=None,
        extract_config=None,
        timeseries=None,
        trade_strategy=None,
        cache_freq='daily',
        label='algo'):
    """build_algo_request

    Create a dictionary for building an algorithm. This is
    opinionated to how the underlying date-based caching
    strategy is running per day. Each business day becomes
    a possible dataset to process with an algorithm.

    :param ticker: ticker
    :param tickers: optional - list of tickers
    :param use_key: redis and s3 to store the algo result
    :param start_date: string date format ``YYYY-MM-DD HH:MM:SS``
    :param end_date: string date format ``YYYY-MM-DD HH:MM:SS``
    :param datasets: list of string dataset types
    :param balance: starting capital balance
    :param commission: commission for buy or sell
    :param num_shares: optional - integer number of starting shares
    :param cache_freq: optional - cache frequency
        (``daily`` is default)
    :param label: optional - algo log tracking name
    :param config_file: path to a json file
        containing custom algorithm object
        member values (like indicator configuration and
        predict future date units ahead for a backtest)
    :param config_dict: optional - dictionary that
        can be passed to derived class implementations
        of: ``def load_from_config(config_dict=config_dict)``

    **Timeseries**

    :param timeseries: optional - string to
        set ``day`` or ``minute`` backtesting
        or live trading
        (default is ``minute``)

    **Trading Strategy**

    :param trade_strategy: optional - string to
        set the type of ``Trading Strategy``
        for backtesting or live trading
        (default is ``count``)

    **Algorithm Dataset Extraction, Loading and Publishing arguments**

    :param load_config: optional - dictionary
        for setting member variables to load an
        agorithm-ready dataset from
        a file, s3 or redis
    :param history_config: optional - dictionary
        for setting member variables to publish
        an algo ``trade history`` to s3, redis, a file
        or slack
    :param report_config: optional - dictionary
        for setting member variables to publish
        an algo ``trading performance report`` to s3,
        redis, a file or slack
    :param extract_config: optional - dictionary
        for setting member variables to publish
        an algo ``trading performance report`` to s3,
        redis, a file or slack

    """
    use_tickers = []
    if ticker:
        use_tickers = [
            ticker.upper()
        ]
    if tickers:
        for t in tickers:
            if t not in use_tickers:
                use_tickers.append(t.upper())

    s3_bucket_name = ae_consts.ALGO_RESULT_S3_BUCKET_NAME
    s3_key = use_key
    redis_key = use_key
    s3_enabled = True
    redis_enabled = True

    work = {
        'tickers': use_tickers,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled,
        'extract_datasets': [],
        'cache_freq': cache_freq,
        'config_file': config_file,
        'config_dict': config_dict,
        'balance': balance,
        'commission': commission,
        'load_config': load_config,
        'history_config': history_config,
        'report_config': report_config,
        'extract_config': extract_config,
        'start_date': None,
        'end_date': None,
        'timeseries': timeseries,
        'trade_strategy': trade_strategy,
        'version': 1,
        'label': label
    }

    start_date_val = ae_utils.get_date_from_str(start_date)
    end_date_val = ae_utils.get_date_from_str(end_date)
    if start_date_val > end_date_val:
        raise Exception(
            'Invalid start_date={} must be less than end_date={}'.format(
                start_date,
                end_date))

    use_dates = []
    new_dataset = None
    cur_date = start_date_val
    if not work['start_date']:
        work['start_date'] = start_date_val.strftime(
            ae_consts.COMMON_TICK_DATE_FORMAT)
    if not work['end_date']:
        work['end_date'] = end_date_val.strftime(
            ae_consts.COMMON_TICK_DATE_FORMAT)
    while cur_date <= end_date_val:
        if cur_date.weekday() < 5:
            for t in use_tickers:
                if cache_freq == 'daily':
                    new_dataset = '{}_{}'.format(
                        t,
                        cur_date.strftime(
                            ae_consts.COMMON_DATE_FORMAT))
                else:
                    new_dataset = '{}_{}'.format(
                        t,
                        cur_date.strftime(
                            ae_consts.COMMON_TICK_DATE_FORMAT))
                if new_dataset:
                    use_dates.append(new_dataset)
                new_dataset = None
            # end for all tickers
        # end of valid days M-F
        if cache_freq == 'daily':
            cur_date += datetime.timedelta(days=1)
        else:
            cur_date += datetime.timedelta(minute=1)
    # end of walking all dates to add

    if len(use_dates) > 0:
        work['extract_datasets'] = use_dates

        log.debug(
            'tickers={} balance={} start={} end={} '
            'cache_freq={} request={}'.format(
                work['tickers'],
                work['balance'],
                work['extract_datasets'][0],
                work['extract_datasets'][-1],
                cache_freq,
                ae_consts.ppj(work)))
    else:
        log.error(
            'there are not enough dates to test between start={} end={} '
            'tickers={} request={}'.format(
                start_date_val,
                end_date_val,
                work['tickers'],
                cache_freq,
                ae_consts.ppj(work)))

    return work
# end of build_algo_request



import os

broker_url = os.getenv(
    'WORKER_BROKER_URL',
    'redis://0.0.0.0:6379/13')
result_backend = os.getenv(
    'WORKER_BACKEND_URL',
    'redis://0.0.0.0:6379/14')

# http://docs.celeryproject.org/en/latest/userguide/optimizing.html

# these are targeted at optimizing processing
# on long-running tasks
# while increasing reliability

# http://docs.celeryproject.org/en/latest/userguide/configuration.html#std:setting-worker_prefetch_multiplier  # noqa
worker_prefetch_multiplier = 1

# http://docs.celeryproject.org/en/latest/userguide/configuration.html#std:setting-broker_heartbeat  # noqa
broker_heartbeat = 240  # seconds

# http://docs.celeryproject.org/en/latest/userguide/configuration.html#std:setting-broker_connection_max_retries  # noqa
broker_connection_max_retries = None

# http://docs.celeryproject.org/en/latest/userguide/configuration.html#std:setting-task_acks_late  # noqa
task_acks_late = True

# http://docs.celeryproject.org/en/latest/userguide/calling.html#calling-retry
task_publish_retry_policy = {
    'interval_max': 1,
    'max_retries': 120,     # None = forever
    'interval_start': 0.1,
    'interval_step': 0.2}

task_serializer = 'json'
result_serializer = 'json'
accept_content = ['json']
timezone = 'America/Los_Angeles'

task_routes = {}



"""
Helper for building a dictionary for the:
``analysis_engine.publish.publish`` function
"""

from analysis_engine.consts import ENABLED_REDIS_PUBLISH
from analysis_engine.consts import REDIS_ADDRESS
from analysis_engine.consts import REDIS_DB
from analysis_engine.consts import REDIS_PASSWORD
from analysis_engine.consts import REDIS_EXPIRE
from analysis_engine.consts import ENABLED_S3_UPLOAD
from analysis_engine.consts import S3_ADDRESS
from analysis_engine.consts import S3_BUCKET
from analysis_engine.consts import S3_ACCESS_KEY
from analysis_engine.consts import S3_SECRET_KEY
from analysis_engine.consts import S3_REGION_NAME
from analysis_engine.consts import S3_SECURE
from analysis_engine.consts import ppj
from spylunking.log.setup_logging import build_colorized_logger

log = build_colorized_logger(
    name=__name__)


def build_publish_request(
        ticker=None,
        tickers=None,
        convert_to_json=False,
        output_file=None,
        compress=False,
        redis_enabled=ENABLED_REDIS_PUBLISH,
        redis_key=None,
        redis_address=REDIS_ADDRESS,
        redis_db=REDIS_DB,
        redis_password=REDIS_PASSWORD,
        redis_expire=REDIS_EXPIRE,
        redis_serializer='json',
        redis_encoding='utf-8',
        s3_enabled=ENABLED_S3_UPLOAD,
        s3_key=None,
        s3_address=S3_ADDRESS,
        s3_bucket=S3_BUCKET,
        s3_access_key=S3_ACCESS_KEY,
        s3_secret_key=S3_SECRET_KEY,
        s3_region_name=S3_REGION_NAME,
        s3_secure=S3_SECURE,
        slack_enabled=False,
        slack_code_block=False,
        slack_full_width=False,
        verbose=False,
        label='publisher'):
    """build_publish_request

    Build a dictionary for helping to quickly publish
    to multiple optional endpoints:
    - a local file path (``output_file``)
    - minio (``s3_bucket`` and ``s3_key``)
    - redis (``redis_key``)
    - slack

    :param ticker: ticker
    :param tickers: optional - list of tickers
    :param label: optional - algo log tracking name
    :param output_file: path to save the data
        to a file
    :param compress: optional - compress before publishing
    :param verbose: optional - boolean to log output
    :param kwargs: optional - future argument support

    **(Optional) Redis connectivity arguments**

    :param redis_enabled: bool - toggle for auto-caching all
        datasets in Redis
        (default is ``ENABLED_REDIS_PUBLISH``)
    :param redis_key: string - key to save the data in redis
        (default is ``None``)
    :param redis_address: Redis connection string format: ``host:port``
        (default is ``REDIS_ADDRESS``)
    :param redis_db: Redis db to use
        (default is ``REDIS_DB``)
    :param redis_password: optional - Redis password
        (default is ``REDIS_PASSWORD``)
    :param redis_expire: optional - Redis expire value
        (default is ``REDIS_EXPIRE``)
    :param redis_serializer: not used yet - support for future
        pickle objects in redis (default is ``json``)
    :param redis_encoding: format of the encoded key in redis
        (default is ``utf-8``)

    **(Optional) Minio (S3) connectivity arguments**

    :param s3_enabled: bool - toggle for auto-archiving on Minio (S3)
        (default is ``ENABLED_S3_UPLOAD``)
    :param s3_key: string - key to save the data in redis
        (default is ``None``)
    :param s3_address: Minio S3 connection string format: ``host:port``
        (default is ``S3_ADDRESS``)
    :param s3_bucket: S3 Bucket for storing the artifacts
        (default is ``S3_BUCKET``) which should be viewable on a browser:
        http://localhost:9000/minio/dev/
    :param s3_access_key: S3 Access key
        (default is ``S3_ACCESS_KEY``)
    :param s3_secret_key: S3 Secret key
        (default is ``S3_SECRET_KEY``)
    :param s3_region_name: S3 region name
        (default is ``S3_REGION_NAME``)
    :param s3_secure: Transmit using tls encryption
        (default is ``S3_SECURE``)

    **(Optional) Slack arguments**

    :param slack_enabled: optional - boolean for
        publishing to slack
    :param slack_code_block: optional - boolean for
        publishing as a code black in slack
    :param slack_full_width: optional - boolean for
        publishing as a to slack using the full
        width allowed
    """
    use_tickers = []
    if ticker:
        use_tickers = [
            ticker.upper()
        ]
    if tickers:
        for t in tickers:
            if t not in use_tickers:
                use_tickers.append(t.upper())

    work = {
        'tickers': use_tickers,
        'label': label,
        'convert_to_json': convert_to_json,
        'output_file': output_file,
        'compress': compress,
        'redis_enabled': redis_enabled,
        'redis_key': redis_key,
        'redis_address': redis_address,
        'redis_db': redis_db,
        'redis_password': redis_password,
        'redis_expire': redis_expire,
        'redis_serializer': redis_serializer,
        'redis_encoding': redis_encoding,
        's3_enabled': s3_enabled,
        's3_key': s3_key,
        's3_address': s3_address,
        's3_bucket': s3_bucket,
        's3_access_key': s3_access_key,
        's3_secret_key': s3_secret_key,
        's3_region_name': s3_region_name,
        's3_secure': s3_secure,
        'slack_enabled': slack_enabled,
        'slack_code_block': slack_code_block,
        'slack_full_width': slack_full_width,
        'verbose': verbose,
        'version': 1,
        'label': label
    }

    log.debug(
        'created publish_request={}'.format(
            ppj(work)))

    return work
# end of build_publish_request



"""
Wrapper for downloading an S3 key as a string
"""

import json
import zlib
from spylunking.log.setup_logging import build_colorized_logger


log = build_colorized_logger(
    name=__name__)


def s3_read_contents_from_key(
        s3,
        s3_bucket_name,
        s3_key,
        encoding='utf-8',
        convert_as_json=True,
        compress=False):
    """s3_read_contents_from_key

    Download the S3 key contents as a string. This
    will raise exceptions.

    :param s3: existing S3 object
    :param s3_bucket_name: bucket name
    :param s3_key: S3 key
    :param encoding: utf-8 by default
    :param convert_to_json: auto-convert to a dict
    :param compress: decompress using ``zlib``
    """

    log.debug(
        'getting s3.Object({}, {})'.format(
            s3_bucket_name,
            s3_key))
    s3_obj = s3.Object(s3_bucket_name, s3_key)

    raw_contents = None
    if compress:
        log.debug(
            'zlib.decompress('
            's3_obj.get()["Body"].read()'
            '.decode({})'
            ''.format(
                encoding))
        raw_contents = zlib.decompress(
            s3_obj.get()['Body'].read()).decode(
                encoding)
    else:
        log.debug(
            's3_obj.get()["Body"].read().decode({})'.format(
                encoding))
        raw_contents = s3_obj.get()['Body'].read().decode(encoding)
    # if compressed or not

    data = None
    if convert_as_json:
        data = json.loads(raw_contents)
    else:
        data = raw_contents
    return data
# end of s3_read_contents_from_key



#!/usr/bin/env python

"""

Fetch new pricing datasets for a single or multiple tickers at once or
pull screeners from FinViz.

1) Fetch pricing data or FinViz screeners results
2) Publish pricing data to redis and minio

"""

import os
import argparse
import celery
import analysis_engine.api_requests as api_requests
import analysis_engine.work_tasks.get_new_pricing_data as task_pricing
import analysis_engine.work_tasks.task_screener_analysis as screener_utils
import spylunking.log.setup_logging as log_utils
import analysis_engine.work_tasks.get_celery_app as get_celery_app
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils


# Disable celery log hijacking
# https://github.com/celery/celery/issues/2509
@celery.signals.setup_logging.connect
def setup_celery_logging(**kwargs):
    pass


log = log_utils.build_colorized_logger(
    name='fetch',
    log_config_path=ae_consts.LOG_CONFIG_PATH)


def start_screener_analysis(
        req):
    """start_screener_analysis

    Start screener-driven analysis with a simple workflow:

    1) Convert FinViz screeners into a list of tickers
       and a ``pandas.DataFrames`` from each ticker's html row
    2) Build unique list of tickers
    3) Pull datasets for each ticker
    4) Run sale-side processing - coming soon
    5) Run buy-side processing - coming soon
    6) Issue alerts to slack - coming soon

    :param req: dictionary to start the screener analysis
    """
    label = req.get(
        'label',
        'screener')
    log.info(
        '{} - start screener analysis'.format(
            label))
    req['celery_disabled'] = True
    analysis_res = screener_utils.run_screener_analysis(
        work_dict=req)
    log.info(
        '{} - done screener analysis result={}'.format(
            label,
            analysis_res))
# end of start_screener_analysis


def fetch_new_stock_datasets():
    """fetch_new_stock_datasets

    Collect all datasets for the ticker **SPY**:

    ::

        fetch_new_stock_datasets.py -t SPY

    .. note:: This requires the following services are listening on:

        - redis ``localhost:6379``
        - minio ``localhost:9000``

    """
    log.info(
        'start - fetch_new_stock_datasets')

    parser = argparse.ArgumentParser(
        description=(
            'Download and store the latest stock pricing, '
            'news, and options chain data '
            'and store it in Minio (S3) and Redis. '
            'Also includes support for getting FinViz '
            'screener tickers'))
    parser.add_argument(
        '-t',
        help=(
            'ticker'),
        required=False,
        dest='ticker')
    parser.add_argument(
        '-g',
        help=(
            'optional - fetch mode: '
            'all = fetch from all data sources (default), '
            'yahoo = fetch from just Yahoo sources, '
            'iex = fetch from just IEX sources'),
        required=False,
        dest='fetch_mode')
    parser.add_argument(
        '-i',
        help=(
            'optional - ticker id '
            'not used without a database'),
        required=False,
        dest='ticker_id')
    parser.add_argument(
        '-e',
        help=(
            'optional - options expiration date'),
        required=False,
        dest='exp_date_str')
    parser.add_argument(
        '-l',
        help=(
            'optional - path to the log config file'),
        required=False,
        dest='log_config_path')
    parser.add_argument(
        '-b',
        help=(
            'optional - broker url for Celery'),
        required=False,
        dest='broker_url')
    parser.add_argument(
        '-B',
        help=(
            'optional - backend url for Celery'),
        required=False,
        dest='backend_url')
    parser.add_argument(
        '-k',
        help=(
            'optional - s3 access key'),
        required=False,
        dest='s3_access_key')
    parser.add_argument(
        '-s',
        help=(
            'optional - s3 secret key'),
        required=False,
        dest='s3_secret_key')
    parser.add_argument(
        '-a',
        help=(
            'optional - s3 address format: <host:port>'),
        required=False,
        dest='s3_address')
    parser.add_argument(
        '-S',
        help=(
            'optional - s3 ssl or not'),
        required=False,
        dest='s3_secure')
    parser.add_argument(
        '-u',
        help=(
            'optional - s3 bucket name'),
        required=False,
        dest='s3_bucket_name')
    parser.add_argument(
        '-G',
        help=(
            'optional - s3 region name'),
        required=False,
        dest='s3_region_name')
    parser.add_argument(
        '-p',
        help=(
            'optional - redis_password'),
        required=False,
        dest='redis_password')
    parser.add_argument(
        '-r',
        help=(
            'optional - redis_address format: <host:port>'),
        required=False,
        dest='redis_address')
    parser.add_argument(
        '-n',
        help=(
            'optional - redis and s3 key name'),
        required=False,
        dest='keyname')
    parser.add_argument(
        '-m',
        help=(
            'optional - redis database number (0 by default)'),
        required=False,
        dest='redis_db')
    parser.add_argument(
        '-x',
        help=(
            'optional - redis expiration in seconds'),
        required=False,
        dest='redis_expire')
    parser.add_argument(
        '-z',
        help=(
            'optional - strike price'),
        required=False,
        dest='strike')
    parser.add_argument(
        '-c',
        help=(
            'optional - contract type "C" for calls "P" for puts'),
        required=False,
        dest='contract_type')
    parser.add_argument(
        '-P',
        help=(
            'optional - get pricing data if "1" or "0" disabled'),
        required=False,
        dest='get_pricing')
    parser.add_argument(
        '-N',
        help=(
            'optional - get news data if "1" or "0" disabled'),
        required=False,
        dest='get_news')
    parser.add_argument(
        '-O',
        help=(
            'optional - get options data if "1" or "0" disabled'),
        required=False,
        dest='get_options')
    parser.add_argument(
        '-U',
        help=(
            'optional - s3 enabled for publishing if "1" or '
            '"0" is disabled'),
        required=False,
        dest='s3_enabled')
    parser.add_argument(
        '-R',
        help=(
            'optional - redis enabled for publishing if "1" or '
            '"0" is disabled'),
        required=False,
        dest='redis_enabled')
    parser.add_argument(
        '-A',
        help=(
            'optional - run an analysis '
            'supported modes: scn'),
        required=False,
        dest='analysis_type')
    parser.add_argument(
        '-L',
        help=(
            'optional - screener urls to pull '
            'tickers for analysis'),
        required=False,
        dest='urls')
    parser.add_argument(
        '-Z',
        help=(
            'disable run without an engine for local testing and demos'),
        required=False,
        dest='celery_enabled',
        action='store_true')
    parser.add_argument(
        '-d',
        help=(
            'debug'),
        required=False,
        dest='debug',
        action='store_true')
    args = parser.parse_args()

    run_offline = True
    ticker = ae_consts.TICKER
    ticker_id = ae_consts.TICKER_ID
    fetch_mode = 'all'
    exp_date_str = ae_consts.NEXT_EXP_STR
    ssl_options = ae_consts.SSL_OPTIONS
    transport_options = ae_consts.TRANSPORT_OPTIONS
    broker_url = ae_consts.WORKER_BROKER_URL
    backend_url = ae_consts.WORKER_BACKEND_URL
    celery_config_module = ae_consts.WORKER_CELERY_CONFIG_MODULE
    include_tasks = ae_consts.INCLUDE_TASKS
    s3_access_key = ae_consts.S3_ACCESS_KEY
    s3_secret_key = ae_consts.S3_SECRET_KEY
    s3_region_name = ae_consts.S3_REGION_NAME
    s3_address = ae_consts.S3_ADDRESS
    s3_secure = ae_consts.S3_SECURE
    s3_bucket_name = ae_consts.S3_BUCKET
    s3_key = ae_consts.S3_KEY
    redis_address = ae_consts.REDIS_ADDRESS
    redis_key = ae_consts.REDIS_KEY
    redis_password = ae_consts.REDIS_PASSWORD
    redis_db = ae_consts.REDIS_DB
    redis_expire = ae_consts.REDIS_EXPIRE
    strike = None
    contract_type = None
    get_pricing = True
    get_news = True
    get_options = True
    s3_enabled = True
    redis_enabled = True
    analysis_type = None
    debug = False

    if args.ticker:
        ticker = args.ticker.upper()
    if args.ticker_id:
        ticker_id = args.ticker_id
    if args.exp_date_str:
        exp_date_str = ae_consts.NEXT_EXP_STR
    if args.broker_url:
        broker_url = args.broker_url
    if args.backend_url:
        backend_url = args.backend_url
    if args.s3_access_key:
        s3_access_key = args.s3_access_key
    if args.s3_secret_key:
        s3_secret_key = args.s3_secret_key
    if args.s3_region_name:
        s3_region_name = args.s3_region_name
    if args.s3_address:
        s3_address = args.s3_address
    if args.s3_secure:
        s3_secure = args.s3_secure
    if args.s3_bucket_name:
        s3_bucket_name = args.s3_bucket_name
    if args.keyname:
        s3_key = args.keyname
        redis_key = args.keyname
    if args.redis_address:
        redis_address = args.redis_address
    if args.redis_password:
        redis_password = args.redis_password
    if args.redis_db:
        redis_db = args.redis_db
    if args.redis_expire:
        redis_expire = args.redis_expire
    if args.strike:
        strike = args.strike
    if args.contract_type:
        contract_type = args.contract_type
    if args.get_pricing:
        get_pricing = args.get_pricing == '1'
    if args.get_news:
        get_news = args.get_news == '1'
    if args.get_options:
        get_options = args.get_options == '1'
    if args.s3_enabled:
        s3_enabled = args.s3_enabled == '1'
    if args.redis_enabled:
        redis_enabled = args.redis_enabled == '1'
    if args.fetch_mode:
        fetch_mode = str(args.fetch_mode).lower()
    if args.analysis_type:
        analysis_type = str(args.analysis_type).lower()
    if args.celery_enabled:
        run_offline = False
    if args.debug:
        debug = True

    work = api_requests.build_get_new_pricing_request()

    work['ticker'] = ticker
    work['ticker_id'] = ticker_id
    work['s3_bucket'] = s3_bucket_name
    work['s3_key'] = s3_key
    work['redis_key'] = redis_key
    work['strike'] = strike
    work['contract'] = contract_type
    work['exp_date'] = exp_date_str
    work['s3_access_key'] = s3_access_key
    work['s3_secret_key'] = s3_secret_key
    work['s3_region_name'] = s3_region_name
    work['s3_address'] = s3_address
    work['s3_secure'] = s3_secure
    work['redis_address'] = redis_address
    work['redis_password'] = redis_password
    work['redis_db'] = redis_db
    work['redis_expire'] = redis_expire
    work['get_pricing'] = get_pricing
    work['get_news'] = get_news
    work['get_options'] = get_options
    work['s3_enabled'] = s3_enabled
    work['redis_enabled'] = redis_enabled
    work['fetch_mode'] = fetch_mode
    work['analysis_type'] = analysis_type
    work['debug'] = debug
    work['label'] = 'ticker={}'.format(
        ticker)

    if analysis_type == 'scn':
        label = 'screener={}'.format(
            work['ticker'])
        fv_urls = []
        if args.urls:
            fv_urls = str(args.urls).split('|')
        if len(fv_urls) == 0:
            fv_urls = os.getenv('SCREENER_URLS', []).split('|')
        screener_req = api_requests.build_screener_analysis_request(
            ticker=ticker,
            fv_urls=fv_urls,
            label=label)
        work.update(screener_req)
        start_screener_analysis(
            req=work)
    # end of analysis_type
    else:
        if not args.keyname:
            last_close_date = ae_utils.last_close()
            work['s3_key'] = '{}_{}'.format(
                work['ticker'],
                last_close_date.strftime(
                    ae_consts.COMMON_DATE_FORMAT))
            work['redis_key'] = '{}_{}'.format(
                work['ticker'],
                last_close_date.strftime(
                    ae_consts.COMMON_DATE_FORMAT))

        path_to_tasks = 'analysis_engine.work_tasks'
        task_name = (
            '{}.get_new_pricing_data.get_new_pricing_data'.format(
                path_to_tasks))
        task_res = None
        if ae_consts.is_celery_disabled() or run_offline:
            work['celery_disabled'] = True
            log.debug(
                'starting without celery work={} offline={}'.format(
                    ae_consts.ppj(work),
                    run_offline))
            task_res = task_pricing.get_new_pricing_data(
                work)

            if debug:
                log.info(
                    'done - result={} '
                    'task={} status={} '
                    'err={} label={}'.format(
                        ae_consts.ppj(task_res),
                        task_name,
                        ae_consts.get_status(status=task_res['status']),
                        task_res['err'],
                        work['label']))
            else:
                log.info(
                    'done - result '
                    'task={} status={} '
                    'err={} label={}'.format(
                        task_name,
                        ae_consts.get_status(status=task_res['status']),
                        task_res['err'],
                        work['label']))
            # if/else debug
        else:
            log.info(
                'connecting to broker={} backend={}'.format(
                    broker_url,
                    backend_url))

            # Get the Celery app
            app = get_celery_app.get_celery_app(
                name=__name__,
                auth_url=broker_url,
                backend_url=backend_url,
                path_to_config_module=celery_config_module,
                ssl_options=ssl_options,
                transport_options=transport_options,
                include_tasks=include_tasks)

            log.info(
                'calling task={} - work={}'.format(
                    task_name,
                    ae_consts.ppj(work)))
            job_id = app.send_task(
                task_name,
                (work,))
            log.info(
                'calling task={} - success job_id={}'.format(
                    task_name,
                    job_id))
        # end of if/else
    # end of supported modes
# end of fetch_new_stock_datasets


if __name__ == '__main__':
    fetch_new_stock_datasets()



"""
Helper for writing to a file - used by unittests
"""

import os


def write_to_file(
        output_file,
        data):
    """write_to_file

    write ``data`` to the ``output_file`` file

    :param output_file: path to file
    :param data: string contents for ``output_file``
    """
    with open(output_file, 'w') as cur_file:
        cur_file.write(data)

    return os.path.exists(output_file)
# end of write_to_file



"""
Perform dataset scrubbing actions
and return the scrubbed dataset as a ready-to-go
data feed. This is an approach for normalizing
an internal data feed.

Supported environment variables:

::

    # verbose logging in this module
    # note this can take longer to transform
    # DataFrames and is not recommended for
    # production:
    export DEBUG_FETCH=1

Ingress Scrubbing supports converting an incoming
dataset (from IEX) and converts it to one
of the following data feed and returned as a
``pandas DataFrame``:

::

    DATAFEED_DAILY = 900
    DATAFEED_MINUTE = 901
    DATAFEED_QUOTE = 902
    DATAFEED_STATS = 903
    DATAFEED_PEERS = 904
    DATAFEED_NEWS = 905
    DATAFEED_FINANCIALS = 906
    DATAFEED_EARNINGS = 907
    DATAFEED_DIVIDENDS = 908
    DATAFEED_COMPANY = 909
    DATAFEED_PRICING_YAHOO = 1100
    DATAFEED_OPTIONS_YAHOO = 1101
    DATAFEED_NEWS_YAHOO = 1102

"""

import datetime
import pandas as pd
import analysis_engine.iex.utils
from spylunking.log.setup_logging import build_colorized_logger
from analysis_engine.consts import ev
from analysis_engine.consts import IEX_DAILY_DATE_FORMAT
from analysis_engine.consts import IEX_MINUTE_DATE_FORMAT
from analysis_engine.consts import COMMON_TICK_DATE_FORMAT
from analysis_engine.consts import IEX_QUOTE_DATE_FORMAT
from analysis_engine.iex.consts import DATAFEED_DAILY
from analysis_engine.iex.consts import DATAFEED_MINUTE
from analysis_engine.iex.consts import DATAFEED_QUOTE
from analysis_engine.iex.consts import DATAFEED_STATS
from analysis_engine.iex.consts import DATAFEED_PEERS
from analysis_engine.iex.consts import DATAFEED_NEWS
from analysis_engine.iex.consts import DATAFEED_FINANCIALS
from analysis_engine.iex.consts import DATAFEED_EARNINGS
from analysis_engine.iex.consts import DATAFEED_DIVIDENDS
from analysis_engine.iex.consts import DATAFEED_COMPANY
from analysis_engine.iex.consts import get_datafeed_str
from analysis_engine.yahoo.consts import DATAFEED_PRICING_YAHOO
from analysis_engine.yahoo.consts import DATAFEED_OPTIONS_YAHOO
from analysis_engine.yahoo.consts import DATAFEED_NEWS_YAHOO
from analysis_engine.yahoo.consts import get_datafeed_str_yahoo

log = build_colorized_logger(
    name=__name__)


def debug_msg(
        label,
        datafeed_type,
        msg_format,
        date_str,
        df):
    """debug_msg

    Debug helper for debugging scrubbing handlers

    :param label: log label
    :param datafeed_type: fetch type
    :param msg_format: message to include
    :param date_str: date string
    :param df: ``pandas DataFrame`` or ``None``
    """

    msg = msg_format.format('_', date_str)

    dft_msg = ''
    if (
            datafeed_type == DATAFEED_PRICING_YAHOO or
            datafeed_type == DATAFEED_OPTIONS_YAHOO or
            datafeed_type == DATAFEED_NEWS_YAHOO):
        dft_msg = get_datafeed_str_yahoo(
            df_type=datafeed_type)
    else:
        dft_msg = get_datafeed_str(
            df_type=datafeed_type)

    if ev('DEBUG_FETCH', '0') == '1':
        if 'START' in msg:
            log.info(
                '{} - {} -------------------------'
                '------------------------------------'.format(
                    label,
                    dft_msg))
        msg = msg_format.format(
            df,
            date_str),
        if hasattr(df, 'empty'):
            log.info(
                '{} - {} - {} found df={} '
                'columns={}'.format(
                    label,
                    dft_msg,
                    msg,
                    df,
                    df.columns.values))
        else:
            log.info(
                '{} - {} - {} not df={}'.format(
                    label,
                    dft_msg,
                    msg,
                    df))

        if 'END' in msg:
            log.info(
                '{} - {} -------------------------'
                '------------------------------------'.format(
                    label,
                    dft_msg))
    else:
        log.info(
            '{} - {} - {}'.format(
                label,
                dft_msg,
                msg))
    # end of debug pre-scrub logging

# end of debug_msg


def build_dates_from_df_col(
        df,
        use_date_str,
        src_col='label',
        src_date_format=IEX_MINUTE_DATE_FORMAT,
        output_date_format=COMMON_TICK_DATE_FORMAT):
    """build_dates_from_df_col

    Converts a string date column series in a ``pandas.DataFrame``
    to a well-formed date string list.

    :param src_col: source column name
    :param use_date_str: date string for today
    :param src_date_format: format of the string in the
                            ```df[src_col]`` columne
    :param output_date_format: write the new date strings
                               in this format.
    :param df: source ``pandas.DataFrame``
    """
    new_dates = []
    for idx, i in enumerate(df['label']):
        org_new_str = ''
        if ':' not in i:
            org_new_str = '{} {}:00:00 {}'.format(
                use_date_str,
                i.split(' ')[0],
                i.split(' ')[1])
        else:
            org_new_str = '{} {}:00 {}'.format(
                use_date_str,
                i.split(' ')[0],
                i.split(' ')[1])
        new_date_val = datetime.datetime.strptime(
            org_new_str,
            src_date_format)
        new_str = new_date_val.strftime(
            output_date_format)
        new_dates.append(new_str)
    # for all rows

    return new_dates
# end of build_dates_from_df_col


def ingress_scrub_dataset(
        label,
        datafeed_type,
        df,
        date_str=None,
        msg_format=None,
        scrub_mode='sort-by-date',
        ds_id='no-id'):
    """ingress_scrub_dataset

    Scrub a ``pandas.DataFrame`` from an Ingress pricing service
    and return the resulting ``pandas.DataFrame``

    :param label: log label
    :param datafeed_type: ``analysis_engine.iex.consts.DATAFEED_*`` type
        or ``analysis_engine.yahoo.consts.DATAFEED_*```
        type
        ::

            DATAFEED_DAILY = 900
            DATAFEED_MINUTE = 901
            DATAFEED_QUOTE = 902
            DATAFEED_STATS = 903
            DATAFEED_PEERS = 904
            DATAFEED_NEWS = 905
            DATAFEED_FINANCIALS = 906
            DATAFEED_EARNINGS = 907
            DATAFEED_DIVIDENDS = 908
            DATAFEED_COMPANY = 909
            DATAFEED_PRICING_YAHOO = 1100
            DATAFEED_OPTIONS_YAHOO = 1101
            DATAFEED_NEWS_YAHOO = 1102
    :param df: ``pandas DataFrame``
    :param date_str: date string for simulating historical dates
                     or ``datetime.datetime.now()`` if not
                     set
    :param msg_format: msg format for a ``string.format()``
    :param scrub_mode: mode to scrub this dataset
    :param ds_id: dataset identifier
    """

    if not hasattr(df, 'empty'):
        log.info(
            '{} - {} no dataset_id={}'.format(
                label,
                datafeed_type,
                ds_id))
        return None

    out_df = df

    daily_date_format = '%I:%M %p'
    minute_date_format = '%I:%M %p'

    use_msg_format = msg_format
    if not msg_format:
        use_msg_format = 'df={} date_str={}'

    use_date_str = date_str
    last_close_date = analysis_engine.iex.utils.last_close()
    today_str = last_close_date.strftime('%Y-%m-%d')

    year_str = today_str.split('-')[0]
    if not use_date_str:
        use_date_str = today_str

    daily_date_format = IEX_DAILY_DATE_FORMAT
    minute_date_format = IEX_MINUTE_DATE_FORMAT

    debug_msg(
        label=label,
        datafeed_type=datafeed_type,
        msg_format='START - {}'.format(
            use_msg_format),
        date_str=use_date_str,
        df=df)

    try:
        if scrub_mode == 'sort-by-date':
            if datafeed_type == DATAFEED_DAILY:
                new_dates = []
                if 'label' in df:
                    for idx, i in enumerate(out_df['label']):
                        new_str = ''
                        if ',' not in i:
                            # Oct 3
                            new_str = '{}-{}-{}'.format(
                                year_str,
                                i.split(' ')[0],
                                i.split(' ')[1])
                        else:
                            # Aug 29, 18
                            new_str = '20{}-{}-{}'.format(
                                i.split(' ')[2],
                                i.split(' ')[0],
                                i.split(' ')[1]).replace(',', '')
                        new_dates.append(new_str)
                    # end for all rows
                    out_df['date'] = pd.to_datetime(
                        new_dates,
                        format=daily_date_format)
                # end if label is in df
            elif datafeed_type == DATAFEED_MINUTE:
                new_dates = []
                if 'label' in df:
                    new_dates = build_dates_from_df_col(
                        src_col='label',
                        src_date_format=minute_date_format,
                        use_date_str=use_date_str,
                        df=out_df)
                    out_df['date'] = pd.to_datetime(
                        new_dates,
                        format='%Y-%m-%d %H:%M:%S')
                # end if label is in df
            elif datafeed_type == DATAFEED_QUOTE:
                columns_list = out_df.columns.values
                if 'latestTime' in columns_list:
                    out_df['date'] = pd.to_datetime(
                        out_df['latestTime'],
                        format=IEX_QUOTE_DATE_FORMAT)
                if 'latestUpdate' in columns_list:
                    out_df['latest_update'] = pd.to_datetime(
                        out_df['latestUpdate'],
                        unit='ns')
                if 'extendedPriceTime' in columns_list:
                    out_df['extended_price_time'] = pd.to_datetime(
                        out_df['extendedPriceTime'],
                        unit='ns')
                if 'iexLastUpdated' in columns_list:
                    out_df['iex_last_update'] = pd.to_datetime(
                        out_df['iexLastUpdated'],
                        unit='ns')
                if 'openTime' in columns_list:
                    out_df['open_time'] = pd.to_datetime(
                        out_df['openTime'],
                        unit='ns')
                if 'closeTime' in columns_list:
                    out_df['close_time'] = pd.to_datetime(
                        out_df['closeTime'],
                        unit='ns')
                # end if label is in df
            elif datafeed_type == DATAFEED_STATS:
                log.info(
                    '{} - {} - no scrub_mode={} '
                    'support'.format(
                        label,
                        datafeed_type,
                        scrub_mode))
                if 'label' in df:
                    out_df['date'] = pd.to_datetime(
                        df['label'],
                        format=daily_date_format)
            elif datafeed_type == DATAFEED_PEERS:
                log.info(
                    '{} - {} - no scrub_mode={} '
                    'support'.format(
                        label,
                        datafeed_type,
                        scrub_mode))
                if 'label' in df:
                    out_df['date'] = pd.to_datetime(
                        df['label'],
                        format=daily_date_format)
            elif datafeed_type == DATAFEED_NEWS:
                log.info(
                    '{} - {} - no scrub_mode={} '
                    'support'.format(
                        label,
                        datafeed_type,
                        scrub_mode))
                if 'label' in df:
                    out_df['date'] = pd.to_datetime(
                        df['label'],
                        format=daily_date_format)
            elif datafeed_type == DATAFEED_FINANCIALS:
                log.info(
                    '{} - {} - no scrub_mode={} '
                    'support'.format(
                        label,
                        datafeed_type,
                        scrub_mode))
                if 'label' in df:
                    out_df['date'] = pd.to_datetime(
                        df['label'],
                        format=daily_date_format)
            elif datafeed_type == DATAFEED_EARNINGS:
                log.info(
                    '{} - {} - no scrub_mode={} '
                    'support'.format(
                        label,
                        datafeed_type,
                        scrub_mode))
                if 'label' in df:
                    out_df['date'] = pd.to_datetime(
                        df['label'],
                        format=daily_date_format)
            elif datafeed_type == DATAFEED_DIVIDENDS:
                log.info(
                    '{} - {} - no scrub_mode={} '
                    'support'.format(
                        label,
                        datafeed_type,
                        scrub_mode))
                if 'label' in df:
                    out_df['date'] = pd.to_datetime(
                        df['label'],
                        format=daily_date_format)
            elif datafeed_type == DATAFEED_COMPANY:
                log.info(
                    '{} - {} - no scrub_mode={} '
                    'support'.format(
                        label,
                        datafeed_type,
                        scrub_mode))
                if 'label' in df:
                    out_df['date'] = pd.to_datetime(
                        df['label'],
                        format=daily_date_format)
            elif datafeed_type == DATAFEED_PRICING_YAHOO:
                log.info(
                    '{} - {} - no scrub_mode={} '
                    'support'.format(
                        label,
                        datafeed_type,
                        scrub_mode))
                if 'date' in df:
                    out_df['date'] = pd.to_datetime(
                        df['date'],
                        format=daily_date_format)
            elif datafeed_type == DATAFEED_OPTIONS_YAHOO:
                log.info(
                    '{} - {} - no scrub_mode={} '
                    'support'.format(
                        label,
                        datafeed_type,
                        scrub_mode))
                if 'date' in df:
                    out_df['date'] = pd.to_datetime(
                        df['date'],
                        format=daily_date_format)
            elif datafeed_type == DATAFEED_NEWS_YAHOO:
                log.info(
                    '{} - {} - no scrub_mode={} '
                    'support'.format(
                        label,
                        datafeed_type,
                        scrub_mode))
                if 'date' in df:
                    out_df['date'] = pd.to_datetime(
                        df['date'],
                        format=daily_date_format)
            else:
                log.info(
                    '{} - {} - no scrub_mode={} '
                    'support'.format(
                        label,
                        datafeed_type,
                        scrub_mode))
                if 'label' in df:
                    out_df['date'] = pd.to_datetime(
                        df['label'],
                        format=daily_date_format)
            # if/else
        else:
            log.info(
                '{} - {} - no scrub_mode'.format(
                    label,
                    datafeed_type))
    except Exception as e:
        log.critical(
            '{} - {} sort={} - '
            'failed with ex={} data={}'.format(
                label,
                datafeed_type,
                scrub_mode,
                e,
                df))
        out_df = None
    # end of try/ex

    debug_msg(
        label=label,
        datafeed_type=datafeed_type,
        msg_format='END - df={} date_str={}',
        date_str=use_date_str,
        df=out_df)

    return out_df
# end of ingress_scrub_dataset


def extract_scrub_dataset(
        label,
        datafeed_type,
        df,
        date_str=None,
        msg_format=None,
        scrub_mode='sort-by-date',
        ds_id='no-id'):
    """extract_scrub_dataset

    Scrub a cached ``pandas.DataFrame`` that was stored
    in Redis and return the resulting ``pandas.DataFrame``

    :param label: log label
    :param datafeed_type: ``analysis_engine.iex.consts.DATAFEED_*`` type
        or ``analysis_engine.yahoo.consts.DATAFEED_*```
        type
        ::

            DATAFEED_DAILY = 900
            DATAFEED_MINUTE = 901
            DATAFEED_QUOTE = 902
            DATAFEED_STATS = 903
            DATAFEED_PEERS = 904
            DATAFEED_NEWS = 905
            DATAFEED_FINANCIALS = 906
            DATAFEED_EARNINGS = 907
            DATAFEED_DIVIDENDS = 908
            DATAFEED_COMPANY = 909
            DATAFEED_PRICING_YAHOO = 1100
            DATAFEED_OPTIONS_YAHOO = 1101
            DATAFEED_NEWS_YAHOO = 1102
    :param df: ``pandas DataFrame``
    :param date_str: date string for simulating historical dates
                     or ``datetime.datetime.now()`` if not
                     set
    :param msg_format: msg format for a ``string.format()``
    :param scrub_mode: mode to scrub this dataset
    :param ds_id: dataset identifier
    """

    if not hasattr(df, 'empty'):
        log.info(
            '{} - {} no dataset_id={}'.format(
                label,
                datafeed_type,
                ds_id))
        return None

    return df
# end of extract_scrub_dataset
